__author__ = "Simon Nilsson"

import io
import itertools
import os
import random
import shutil
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import imutils
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import PIL
import seaborn as sns
from matplotlib import cm
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from numba import njit

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import simba
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.enums import Formats, Options, TextOptions
from simba.utils.lookups import get_color_dict, get_named_colors
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_frm_of_video


class PlottingMixin(object):
    """
    Methods for visualizations
    """

    def __init__(self):
        pass

    def create_gantt_img(
        self,
        bouts_df: pd.DataFrame,
        clf_name: str,
        image_index: int,
        fps: int,
        gantt_img_title: str,
    ):
        """
        Helper to create a single gantt plot based on the data preceeding the input image

        :param pd.DataFrame bouts_df: ataframe holding information on individual bouts created by :meth:`simba.misc_tools.get_bouts_for_gantt`.
        :param str clf_name: Name of the classifier.
        :param int image_index: The count of the image. E.g., ``1000`` will create a gantt image representing frame 1-1000.
        :param int fps: The fps of the input video.
        :param str gantt_img_title: Title of the image.
        :return np.ndarray
        """

        fig, ax = plt.subplots()
        fig.suptitle(gantt_img_title, fontsize=24)
        relRows = bouts_df.loc[bouts_df["End_frame"] <= image_index]
        for i, event in enumerate(relRows.groupby("Event")):
            data_event = event[1][["Start_time", "Bout_time"]]
            ax.broken_barh(data_event.values, (4, 4), facecolors="red")
        xLength = (round(image_index / fps)) + 1
        if xLength < 10:
            xLength = 10
        ax.set_xlim(0, xLength)
        ax.set_ylim([0, 12])
        plt.ylabel(clf_name, fontsize=12)
        plt.yticks([])
        plt.xlabel("time(s)", fontsize=12)
        ax.yaxis.set_ticklabels([])
        ax.grid(True)
        buffer_ = io.BytesIO()
        plt.savefig(buffer_, format="png")
        buffer_.seek(0)
        image = PIL.Image.open(buffer_)
        ar = np.asarray(image)
        open_cv_image = ar[:, :, ::-1]
        open_cv_image = cv2.cvtColor(open_cv_image, cv2.COLOR_RGB2BGR)
        open_cv_image = cv2.resize(open_cv_image, (640, 480))
        open_cv_image = np.uint8(open_cv_image)
        buffer_.close()
        plt.close(fig)

        return open_cv_image

    def create_single_color_lst(
        self,
        pallete_name: Literal[Options.PALETTE_OPTIONS],
        increments: int,
        as_rgb_ratio: bool = False,
        as_hex: bool = False,
    ) -> List[Union[str, int, float]]:
        """
        Helper to create a color palette of bgr colors in a list.

        :param str pallete_name: Palette name (e.g., 'jet')
        :param int increments: Numbers of colors in the color palette to create.
        :param bool as_rgb_ratio: If True returns the colors as RGB ratios (0-1).
        :param bool as_hex: If True, returns the colors as HEX.
        :return list

        .. note::
           If as_rgb_ratio **AND** as_hex, then returns HEX.


        """
        if as_hex:
            as_rgb_ratio = True
        cmap = cm.get_cmap(pallete_name, increments + 1)
        color_lst = []
        for i in range(cmap.N):
            rgb = list((cmap(i)[:3]))
            if not as_rgb_ratio:
                rgb = [i * 255 for i in rgb]
            rgb.reverse()
            if as_hex:
                rgb = matplotlib.colors.to_hex(rgb)
            color_lst.append(rgb)
        return color_lst

    def remove_a_folder(self, folder_dir: str) -> None:
        """Helper to remove a directory, use for cleaning up smaller multiprocessed videos following concat"""
        shutil.rmtree(folder_dir, ignore_errors=True)

    def split_and_group_df(
        self,
        df: pd.DataFrame,
        splits: int,
        include_row_index: bool = False,
        include_split_order: bool = True,
    ) -> (List[pd.DataFrame], int):
        """
        Helper to split a dataframe for multiprocessing. If include_split_order, then include the group number
        in split data as a column. If include_row_index, includes a column representing the row index in the array,
        which can be helpful for knowing the frame indexes while multiprocessing videos. Returns split data and approximations of number of observations per split.
        """

        if include_row_index:
            row_indices = np.arange(len(df)).reshape(-1, 1)
            df = np.concatenate((df, row_indices), axis=1)
        data_arr = np.array_split(df, splits)
        if include_split_order:
            for df_cnt in range(len(data_arr)):
                data_arr[df_cnt]["group"] = df_cnt
        obs_per_split = len(data_arr[0])

        return data_arr, obs_per_split

    def make_distance_plot(
        self,
        data: np.array,
        line_attr: Dict[int, str],
        style_attr: Dict[str, Any],
        fps: int,
        save_img: bool = False,
        save_path: Optional[str] = None,
    ) -> np.ndarray:
        """
        Helper to make a single line plot .png image with N lines.

        :param np.array data: Two-dimensional array where rows represent frames and columns represent intertwined x and y coordinates.
        :param dict line_attr: Line color attributes.
        :param dict style_attr: Plot attributes (size, font size, line width etc).
        :param int fps: Video frame rate.
        :param Optionan[str] save_path: Location to store output .png image. If None, then return image.

        .. note::
           `GitHub tutorial/documentation <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-distance-plots>`__.

        :example:
        >>> fps = 10
        >>> data = np.random.random((100,2))
        >>> line_attr = {0: ['Blue'], 1: ['Red']}
        >>> save_path = '/_tests/final_frm.png'
        >>> style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8, 'y_max': 'auto'}
        >>> self.make_distance_plot(fps=fps, data=data, line_attr=line_attr, style_attr=style_attr, save_path=save_path)
        """
        colors = get_color_dict()
        for j in range(data.shape[1]):
            color = colors[line_attr[j][-1]][::-1]
            color = tuple(x / 255 for x in color)
            plt.plot(
                data[:, j],
                color=color,
                linewidth=style_attr["line width"],
                alpha=style_attr["opacity"],
            )

        timer = SimbaTimer(start=True)
        max_x = len(data)
        if style_attr["y_max"] == "auto":
            max_y = np.amax(data)
        else:
            max_y = float(style_attr["y_max"])
        y_ticks_locs = y_lbls = np.round(np.linspace(0, max_y, 10), 2)
        x_ticks_locs = x_lbls = np.linspace(0, max_x, 5)
        x_lbls = np.round((x_lbls / fps), 1)

        plt.xlabel("time (s)")
        plt.ylabel("distance (cm)")
        plt.xticks(
            x_ticks_locs,
            x_lbls,
            rotation="horizontal",
            fontsize=style_attr["font size"],
        )
        plt.yticks(y_ticks_locs, y_lbls, fontsize=style_attr["font size"])
        plt.ylim(0, max_y)
        plt.suptitle(
            "Animal distances", x=0.5, y=0.92, fontsize=style_attr["font size"] + 4
        )
        buffer_ = io.BytesIO()
        plt.savefig(buffer_, format="png")
        buffer_.seek(0)
        img = PIL.Image.open(buffer_)
        img = np.uint8(cv2.cvtColor(np.asarray(img), cv2.COLOR_RGB2BGR))
        buffer_.close()
        plt.close()
        img = cv2.resize(img, (style_attr["width"], style_attr["height"]))
        timer.stop_timer()
        if save_img:
            cv2.imwrite(save_path, img)
            stdout_success(
                f"Final distance plot saved at {save_path}",
                elapsed_time=timer.elapsed_time_str,
                source=self.__class__.__name__,
            )
        else:
            return img

    def make_probability_plot(
        self, data: pd.Series, style_attr: dict, clf_name: str, fps: int, save_path: str
    ) -> np.ndarray:
        """
        Make a single classifier probability plot png image.

        :param pd.Series data: row representing frames and field representing classification probabilities.
        :param dict line_attr: Line color attributes.
        :param dict style_attr: Image attributes (size, font size, line width etc).
        :param int fps: Video frame rate.
        :param str ot
        :param str save_path: Location to store output .png image.



        .. notes:
          `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-classification-probabilities>`__.

        :example:
        >>> data = pd.Series(np.random.random((100, 1)).flatten())
        >>> style_attr = {'width': 640, 'height': 480, 'font size': 10, 'line width': 6, 'color': 'blue', 'circle size': 20}
        >>> clf_name='Attack'
        >>> fps=10
        >>> save_path = '/_test/frames/output/probability_plots/Together_1_final_frame.png'

        >>> _ = self.make_probability_plot(data=data, style_attr=style_attr, clf_name=clf_name, fps=fps, save_path=save_path)
        """

        timer = SimbaTimer()
        timer.start_timer()
        if style_attr["y_max"] == "auto":
            max_y = float(data.max().round(2))
        else:
            max_y = float(style_attr["y_max"])
        max_x = len(data)
        plt.plot(
            list(data), color=style_attr["color"], linewidth=style_attr["line width"]
        )
        plt.plot(
            len(data),
            list(data)[-1],
            "o",
            markersize=style_attr["circle size"],
            color=style_attr["color"],
        )
        plt.ylim([0, max_y])
        plt.ylabel(clf_name, fontsize=style_attr["font size"])

        y_ticks_locs = y_lbls = np.round(np.linspace(0, max_y, 10), 2)
        x_ticks_locs = x_lbls = np.linspace(0, max_x, 5)
        x_lbls = np.round((x_lbls / fps), 1)
        plt.xlabel("Time (s)", fontsize=style_attr["font size"] + 4)
        plt.grid()
        plt.xticks(
            x_ticks_locs,
            x_lbls,
            rotation="horizontal",
            fontsize=style_attr["font size"],
        )
        plt.yticks(y_ticks_locs, y_lbls, fontsize=style_attr["font size"])
        plt.suptitle(
            "{} {}".format(clf_name, "probability"),
            x=0.5,
            y=0.92,
            fontsize=style_attr["font size"] + 4,
        )
        buffer_ = io.BytesIO()
        plt.savefig(buffer_, format="png")
        buffer_.seek(0)
        image = PIL.Image.open(buffer_)
        ar = np.asarray(image)
        img = cv2.cvtColor(ar, cv2.COLOR_RGB2BGR)
        img = np.uint8(cv2.resize(img, (style_attr["width"], style_attr["height"])))
        buffer_.close()
        plt.close()
        timer.stop_timer()
        cv2.imwrite(save_path, img)
        stdout_success(
            msg=f"Final distance plot saved at {save_path}",
            elapsed_time=timer.elapsed_time_str,
            source=self.__class__.__name__,
        )

    def make_path_plot(
        self,
        data_df: pd.DataFrame,
        video_info: pd.DataFrame,
        style_attr: dict,
        deque_dict: dict,
        clf_attr: dict,
        save_path: str,
        print_animal_names: bool = True,
    ) -> None:
        """
        Helper to make a path plot.

        :param pd.DataFrame data_df: Dataframe holding body-part coordinates
        :param pd.DataFrame video_info: Video info dataframe (parsed project_folder/logs/video_info.csv)
        :param dict style_attr: Dict holding image style attributes. E.g., {'width': 'As input', 'height': 'As input', 'line width': 5, 'font size': 5, 'font thickness': 2, 'circle size': 5, 'bg color': 'White', 'max lines': 100}
        :param dict deque_dict: Dict with deque holsing all paths to visualize
        :param dict clf_attr: Dict holding image classifictaion attributes e.g., {0: ['Attack', 'Black', 'Size: 30'], 1: ['Sniffing', 'Red', 'Size: 30']}
        :param str save_path: Location to save image
        """

        video_timer = SimbaTimer(start=True)
        for frm_cnt in range(len(data_df)):
            for animal_cnt, (animal_name, animal_data) in enumerate(deque_dict.items()):
                bp_x = int(data_df.iloc[frm_cnt][f'{animal_data["bp"]}_{"x"}'])
                bp_y = int(data_df.iloc[frm_cnt][f'{animal_data["bp"]}_{"y"}'])
                deque_dict[animal_name]["deque"].appendleft((bp_x, bp_y))

        font = cv2.FONT_HERSHEY_COMPLEX
        color_dict = get_color_dict()

        if not isinstance(style_attr["bg color"], np.ndarray):
            img = np.zeros(
                (
                    int(video_info["Resolution_height"].values[0]),
                    int(video_info["Resolution_width"].values[0]),
                    3,
                )
            )
        else:
            img = np.zeros(
                (
                    int(style_attr["bg color"].shape[0]),
                    int(style_attr["bg color"].shape[1]),
                    3,
                )
            )
        img[:] = style_attr["bg color"]

        for animal_name, animal_data in deque_dict.items():
            cv2.circle(
                img,
                (deque_dict[animal_name]["deque"][0]),
                0,
                deque_dict[animal_name]["clr"],
                style_attr["circle size"],
            )
            if print_animal_names:
                cv2.putText(
                    img,
                    animal_name,
                    (deque_dict[animal_name]["deque"][0]),
                    font,
                    style_attr["font size"],
                    deque_dict[animal_name]["clr"],
                    style_attr["font thickness"],
                )

        for animal_name, animal_data in deque_dict.items():
            for i in range(len(deque_dict[animal_name]["deque"]) - 1):
                line_clr = deque_dict[animal_name]["clr"]
                position_1 = deque_dict[animal_name]["deque"][i]
                position_2 = deque_dict[animal_name]["deque"][i + 1]
                cv2.line(
                    img, position_1, position_2, line_clr, style_attr["line width"]
                )

        if clf_attr:
            for clf_cnt, clf_name in enumerate(clf_attr["data"].columns):
                clf_size = int(clf_attr["attr"][clf_cnt][-1].split(": ")[-1])
                clf_clr = color_dict[clf_attr["attr"][clf_cnt][1]]
                clf_sliced_idx = list(
                    clf_attr["data"][clf_attr["data"][clf_name] == 1].index
                )
                positions = (
                    clf_attr["positions"]
                    .loc[clf_sliced_idx]
                    .reset_index(drop=True)
                    .values
                )
                for i in range(positions.shape[0]):
                    cv2.circle(
                        img, (positions[i][0], positions[i][1]), 0, clf_clr, clf_size
                    )

        video_timer.stop_timer()
        img = cv2.resize(img, (style_attr["width"], style_attr["height"]))
        cv2.imwrite(save_path, img)
        stdout_success(
            msg=f"Final path plot saved at {save_path}",
            elapsed_time=video_timer.elapsed_time_str,
            source=self.__class__.__name__,
        )

    def make_gantt_plot(
        self,
        data_df: pd.DataFrame,
        bouts_df: pd.DataFrame,
        clf_names: List[str],
        fps: int,
        style_attr: dict,
        video_name: str,
        save_path: str,
    ) -> None:

        video_timer = SimbaTimer(start=True)
        colours = get_named_colors()
        colour_tuple_x = list(np.arange(3.5, 203.5, 5))
        fig, ax = plt.subplots()
        for i, event in enumerate(bouts_df.groupby("Event")):
            for x in clf_names:
                if event[0] == x:
                    ix = clf_names.index(x)
                    data_event = event[1][["Start_time", "Bout_time"]]
                    ax.broken_barh(
                        data_event.values,
                        (colour_tuple_x[ix], 3),
                        facecolors=colours[ix],
                    )

        x_ticks_locs = x_lbls = np.round(np.linspace(0, len(data_df) / fps, 6))
        ax.set_xticks(x_ticks_locs)
        ax.set_xticklabels(x_lbls)
        ax.set_ylim(0, colour_tuple_x[len(clf_names)])
        ax.set_yticks(np.arange(5, 5 * len(clf_names) + 1, 5))
        ax.set_yticklabels(clf_names, rotation=style_attr["font rotation"])
        ax.tick_params(axis="both", labelsize=style_attr["font size"])
        plt.xlabel("Session (s)", fontsize=style_attr["font size"] + 3)
        ax.yaxis.grid(True)
        buffer_ = io.BytesIO()
        plt.savefig(buffer_, format="png")
        buffer_.seek(0)
        image = PIL.Image.open(buffer_)
        ar = np.asarray(image)
        open_cv_image = cv2.cvtColor(ar, cv2.COLOR_RGB2BGR)
        open_cv_image = cv2.resize(
            open_cv_image, (style_attr["width"], style_attr["height"])
        )
        frame = np.uint8(open_cv_image)
        buffer_.close()
        plt.close(fig)
        cv2.imwrite(save_path, frame)
        video_timer.stop_timer()
        stdout_success(
            msg=f"Final gantt frame for video {video_name} saved at {save_path}",
            elapsed_time=video_timer.elapsed_time_str,
            source=self.__class__.__name__,
        )

    @staticmethod
    def make_clf_heatmap_plot(
        frm_data: np.array,
        max_scale: float,
        palette: Literal[Options.PALETTE_OPTIONS],
        aspect_ratio: float,
        shading: Literal["gouraud", "flat"],
        clf_name: str,
        img_size: Tuple[int, int],
        file_name: Optional[str] = None,
        final_img: bool = False,
    ):

        cum_df = pd.DataFrame(frm_data).reset_index()
        cum_df = cum_df.melt(
            id_vars="index",
            value_vars=None,
            var_name=None,
            value_name="seconds",
            col_level=None,
        ).rename(columns={"index": "vertical_idx", "variable": "horizontal_idx"})
        cum_df["color"] = (
            (cum_df["seconds"].astype(float) / float(max_scale))
            .round(2)
            .clip(upper=100)
        )
        color_array = np.zeros(
            (
                len(cum_df["vertical_idx"].unique()),
                len(cum_df["horizontal_idx"].unique()),
            )
        )
        for i in range(color_array.shape[0]):
            for j in range(color_array.shape[1]):
                value = cum_df["color"][
                    (cum_df["horizontal_idx"] == j) & (cum_df["vertical_idx"] == i)
                ].values[0]
                color_array[i, j] = value
        color_array = color_array * max_scale
        matplotlib.font_manager._get_font.cache_clear()
        plt.close("all")
        fig = plt.figure()
        im_ratio = color_array.shape[0] / color_array.shape[1]
        plt.pcolormesh(
            color_array,
            shading=shading,
            cmap=palette,
            rasterized=True,
            alpha=1,
            vmin=0.0,
            vmax=float(max_scale),
        )
        plt.gca().invert_yaxis()
        plt.xticks([])
        plt.yticks([])
        plt.axis("off")
        plt.tick_params(axis="both", which="both", length=0)
        cb = plt.colorbar(pad=0.0, fraction=0.023 * im_ratio)
        cb.ax.tick_params(size=0)
        cb.outline.set_visible(False)
        cb.set_label("{} (seconds)".format(clf_name), rotation=270, labelpad=10)
        plt.tight_layout()
        # plt.gca().set_aspect(aspect_ratio)
        buffer_ = io.BytesIO()
        plt.savefig(buffer_, format="png")
        buffer_.seek(0)
        image = PIL.Image.open(buffer_)
        image = np.uint8(cv2.cvtColor(np.asarray(image), cv2.COLOR_RGB2BGR))
        image = cv2.resize(image, img_size)
        buffer_.close()
        plt.close()
        if final_img:
            cv2.imwrite(file_name, image)
            stdout_success(
                msg=f"Final classifier heatmap image saved at at {file_name}",
                source="make_clf_heatmap_plot",
            )
        else:
            return image

    @staticmethod
    def make_location_heatmap_plot(
        frm_data: np.array,
        max_scale: float,
        palette: Literal[Options.PALETTE_OPTIONS],
        aspect_ratio: float,
        shading: str,
        img_size: Tuple[int, int],
        file_name: str or None = None,
        final_img: bool = False,
    ):

        cum_df = pd.DataFrame(frm_data).reset_index()
        cum_df = cum_df.melt(
            id_vars="index",
            value_vars=None,
            var_name=None,
            value_name="seconds",
            col_level=None,
        ).rename(columns={"index": "vertical_idx", "variable": "horizontal_idx"})
        cum_df["color"] = (
            (cum_df["seconds"].astype(float) / float(max_scale))
            .round(2)
            .clip(upper=100)
        )
        color_array = np.zeros(
            (
                len(cum_df["vertical_idx"].unique()),
                len(cum_df["horizontal_idx"].unique()),
            )
        )
        for i in range(color_array.shape[0]):
            for j in range(color_array.shape[1]):
                value = cum_df["color"][
                    (cum_df["horizontal_idx"] == j) & (cum_df["vertical_idx"] == i)
                ].values[0]
                color_array[i, j] = value
        color_array = color_array * max_scale
        matplotlib.font_manager._get_font.cache_clear()
        plt.close("all")
        fig = plt.figure()
        im_ratio = color_array.shape[0] / color_array.shape[1]
        plt.pcolormesh(
            color_array,
            shading=shading,
            cmap=palette,
            rasterized=True,
            alpha=1,
            vmin=0.0,
            vmax=max_scale,
        )
        plt.gca().invert_yaxis()
        plt.xticks([])
        plt.yticks([])
        plt.axis("off")
        plt.tick_params(axis="both", which="both", length=0)
        cb = plt.colorbar(pad=0.0, fraction=0.023 * im_ratio)
        cb.ax.tick_params(size=0)
        cb.outline.set_visible(False)
        cb.set_label("location (seconds)", rotation=270, labelpad=10)
        plt.tight_layout()
        # plt.gca().set_aspect(aspect_ratio)
        canvas = FigureCanvas(fig)
        canvas.draw()
        mat = np.array(canvas.renderer._renderer)
        image = cv2.cvtColor(mat, cv2.COLOR_RGB2BGR)
        image = cv2.resize(image, img_size)
        image = np.uint8(image)
        plt.close("all")
        if final_img:
            cv2.imwrite(file_name, image)
            stdout_success(
                msg=f"Final location heatmap image saved at at {file_name}",
                source=PlottingMixin.make_location_heatmap_plot.__class__.__name__,
            )
        else:
            return image
        return None

    def get_bouts_for_gantt(
        self, data_df: pd.DataFrame, clf_name: str, fps: int
    ) -> np.ndarray:
        """
        Helper to detect all behavior bouts for a specific classifier.

        :param pd.DataFrame data_df: Pandas Dataframe with classifier prediction data.
        :param str clf_name: Name of the classifier
        :param int fps: The fps of the input video.
        :return  pd.DataFrame: Holding the start time, end time, end frame, bout time etc of each classified bout.
        """

        boutsList, nameList, startTimeList, endTimeList, endFrameList = (
            [],
            [],
            [],
            [],
            [],
        )
        groupDf = pd.DataFrame()
        v = (data_df[clf_name] != data_df[clf_name].shift()).cumsum()
        u = data_df.groupby(v)[clf_name].agg(["all", "count"])
        m = u["all"] & u["count"].ge(1)
        groupDf["groups"] = data_df.groupby(v).apply(
            lambda x: (x.index[0], x.index[-1])
        )[m]
        for indexes, rows in groupDf.iterrows():
            currBout = list(rows["groups"])
            boutTime = ((currBout[-1] - currBout[0]) + 1) / fps
            startTime = (currBout[0] + 1) / fps
            endTime = (currBout[1]) / fps
            endFrame = currBout[1]
            endTimeList.append(endTime)
            startTimeList.append(startTime)
            boutsList.append(boutTime)
            nameList.append(clf_name)
            endFrameList.append(endFrame)

        return pd.DataFrame(
            list(zip(nameList, startTimeList, endTimeList, endFrameList, boutsList)),
            columns=["Event", "Start_time", "End Time", "End_frame", "Bout_time"],
        )

    def resize_gantt(self, gantt_img: np.array, img_height: int) -> np.ndarray:
        """
        Helper to resize image while retaining aspect ratio.
        """

        return imutils.resize(gantt_img, height=img_height)

    @staticmethod
    def roi_feature_visualizer_mp(
        data: pd.DataFrame,
        text_locations: dict,
        scalers: dict,
        save_temp_dir: str,
        video_meta_data: dict,
        shape_info: dict,
        style_attr: dict,
        directing_viable: bool,
        video_path: str,
        animal_names: list,
        tracked_bps: dict,
        animal_bps: dict,
        directing_data: pd.DataFrame,
    ):

        fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        font = cv2.FONT_HERSHEY_COMPLEX

        def __insert_texts(shape_info: dict, img: np.array):
            for shape_name, shape_info in shape_info.items():
                for animal_name in animal_names:
                    shape_color = shape_info["Color BGR"]
                    cv2.putText(
                        img,
                        text_locations[animal_name][shape_name]["in_zone_text"],
                        text_locations[animal_name][shape_name]["in_zone_text_loc"],
                        font,
                        scalers["font_size"],
                        shape_color,
                        1,
                    )
                    cv2.putText(
                        img,
                        text_locations[animal_name][shape_name]["distance_text"],
                        text_locations[animal_name][shape_name]["distance_text_loc"],
                        font,
                        scalers["font_size"],
                        shape_color,
                        1,
                    )
                    if directing_viable and style_attr["Directionality"]:
                        cv2.putText(
                            img,
                            text_locations[animal_name][shape_name]["directing_text"],
                            text_locations[animal_name][shape_name][
                                "directing_text_loc"
                            ],
                            font,
                            scalers["font_size"],
                            shape_color,
                            1,
                        )
            return img

        def __insert_shapes(img: np.array, shape_info: dict):
            for shape_name, shape_info in shape_info.items():
                if shape_info["Shape_type"] == "Rectangle":
                    cv2.rectangle(
                        img,
                        (int(shape_info["topLeftX"]), int(shape_info["topLeftY"])),
                        (
                            int(shape_info["Bottom_right_X"]),
                            int(shape_info["Bottom_right_Y"]),
                        ),
                        shape_info["Color BGR"],
                        int(shape_info["Thickness"]),
                    )
                    if style_attr["ROI_centers"]:
                        center_cord = (
                            (int(shape_info["topLeftX"] + (shape_info["width"] / 2))),
                            (int(shape_info["topLeftY"] + (shape_info["height"] / 2))),
                        )
                        cv2.circle(
                            img,
                            center_cord,
                            scalers["circle_size"],
                            shape_info["Color BGR"],
                            -1,
                        )
                    if style_attr["ROI_ear_tags"]:
                        for tag_data in shape_info["Tags"].values():
                            cv2.circle(
                                img,
                                tag_data,
                                scalers["circle_size"],
                                shape_info["Color BGR"],
                                -1,
                            )

                if shape_info["Shape_type"] == "Circle":
                    cv2.circle(
                        img,
                        (int(shape_info["centerX"]), int(shape_info["centerY"])),
                        shape_info["radius"],
                        shape_info["Color BGR"],
                        int(shape_info["Thickness"]),
                    )
                    if style_attr["ROI_centers"]:
                        cv2.circle(
                            img,
                            (int(shape_info["centerX"]), int(shape_info["centerY"])),
                            scalers["circle_size"],
                            shape_info["Color BGR"],
                            -1,
                        )
                    if style_attr["ROI_ear_tags"]:
                        for tag_data in shape_info["Tags"].values():
                            cv2.circle(
                                img,
                                tag_data,
                                scalers["circle_size"],
                                shape_info["Color BGR"],
                                -1,
                            )

                if shape_info["Shape_type"] == "Polygon":
                    cv2.polylines(
                        img,
                        [shape_info["vertices"]],
                        True,
                        shape_info["Color BGR"],
                        thickness=int(shape_info["Thickness"]),
                    )
                    if style_attr["ROI_centers"]:
                        cv2.circle(
                            img,
                            shape_info["Tags"]["Center_tag"],
                            scalers["circle_size"],
                            shape_info["Color BGR"],
                            -1,
                        )
                    if style_attr["ROI_ear_tags"]:
                        for tag_data in shape_info["Tags"].values():
                            cv2.circle(
                                img,
                                tuple(tag_data),
                                scalers["circle_size"],
                                shape_info["Color BGR"],
                                -1,
                            )

            return img

        def __insert_directing_line(
            directing_data: pd.DataFrame,
            shape_name: str,
            frame_cnt: int,
            shape_info: dict,
            img: np.array,
            video_name: str,
            style_attr: dict,
        ):

            r = directing_data.loc[
                (directing_data["Video"] == video_name)
                & (directing_data["ROI"] == shape_name)
                & (directing_data["Animal"] == animal_name)
                & (directing_data["Frame"] == frame_cnt)
            ]
            clr = shape_info[shape_name]["Color BGR"]
            thickness = shape_info[shape_name]["Thickness"]
            if style_attr["Directionality_style"] == "Funnel":
                convex_hull_arr = (
                    np.array(
                        [
                            [r["ROI_edge_1_x"], r["ROI_edge_1_y"]],
                            [r["ROI_edge_2_x"], r["ROI_edge_2_y"]],
                            [r["Eye_x"], r["Eye_y"]],
                        ]
                    )
                    .reshape(-1, 2)
                    .astype(int)
                )
                cv2.fillPoly(img, [convex_hull_arr], clr)

            if style_attr["Directionality_style"] == "Lines":
                cv2.line(
                    img,
                    (int(r["Eye_x"]), int(r["Eye_y"])),
                    (int(r["ROI_x"]), int(r["ROI_y"])),
                    clr,
                    int(thickness),
                )

            return img

        group_cnt = int(data["group"].values[0])
        start_frm, current_frm, end_frm = data.index[0], data.index[0], data.index[-1]
        save_path = os.path.join(save_temp_dir, "{}.mp4".format(str(group_cnt)))
        _, video_name, _ = get_fn_ext(filepath=video_path)
        writer = cv2.VideoWriter(
            save_path,
            fourcc,
            video_meta_data["fps"],
            (video_meta_data["width"] * 2, video_meta_data["height"]),
        )

        cap = cv2.VideoCapture(video_path)
        cap.set(1, start_frm)

        while current_frm < end_frm:
            ret, img = cap.read()
            img = cv2.copyMakeBorder(
                img,
                0,
                0,
                0,
                int(video_meta_data["width"]),
                borderType=cv2.BORDER_CONSTANT,
                value=style_attr["Border_color"],
            )
            img = __insert_texts(shape_info=shape_info, img=img)
            if style_attr["Pose_estimation"]:
                for animal, animal_bp_name in tracked_bps.items():
                    bp_cords = data.loc[current_frm, animal_bp_name].values
                    cv2.circle(
                        img,
                        (int(bp_cords[0]), int(bp_cords[1])),
                        0,
                        animal_bps[animal]["colors"][0],
                        scalers["circle_size"],
                    )
                    cv2.putText(
                        img,
                        animal,
                        (int(bp_cords[0]), int(bp_cords[1])),
                        font,
                        scalers["font_size"],
                        animal_bps[animal]["colors"][0],
                        1,
                    )

            img = __insert_shapes(img=img, shape_info=shape_info)
            #
            for animal_name, shape_name in itertools.product(animal_names, shape_info):
                in_zone_col_name = f'{shape_name} {animal_name} {"in zone"}'
                distance_col_name = f'{shape_name} {animal_name} {"distance"}'
                in_zone_value = str(bool(data.loc[current_frm, in_zone_col_name]))
                distance_value = str(round(data.loc[current_frm, distance_col_name], 2))
                cv2.putText(
                    img,
                    in_zone_value,
                    text_locations[animal_name][shape_name]["in_zone_data_loc"],
                    font,
                    scalers["font_size"],
                    shape_info[shape_name]["Color BGR"],
                    1,
                )
                cv2.putText(
                    img,
                    distance_value,
                    text_locations[animal_name][shape_name]["distance_data_loc"],
                    font,
                    scalers["font_size"],
                    shape_info[shape_name]["Color BGR"],
                    1,
                )
                if directing_viable and style_attr["Directionality"]:
                    facing_col_name = "{} {} {}".format(
                        shape_name, animal_name, "facing"
                    )
                    facing_value = bool(data.loc[current_frm, facing_col_name])
                    cv2.putText(
                        img,
                        str(facing_value),
                        text_locations[animal_name][shape_name]["directing_data_loc"],
                        font,
                        scalers["font_size"],
                        shape_info[shape_name]["Color BGR"],
                        1,
                    )
                    if facing_value:
                        img = __insert_directing_line(
                            directing_data=directing_data,
                            shape_name=shape_name,
                            frame_cnt=current_frm,
                            shape_info=shape_info,
                            img=img,
                            video_name=video_name,
                            style_attr=style_attr,
                        )
            writer.write(img)
            current_frm += 1
            print(
                "Multi-processing video frame {} on core {}...".format(
                    str(current_frm), str(group_cnt)
                )
            )
        cap.release()
        writer.release()

        return group_cnt

    @staticmethod
    def directing_animals_mp(
        data: pd.DataFrame,
        directionality_data: pd.DataFrame,
        bp_names: dict,
        style_attr: dict,
        save_temp_dir: str,
        video_path: str,
        video_meta_data: dict,
        colors: list,
    ):

        fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        group_cnt = int(data.iloc[0]["group"])
        start_frm, current_frm, end_frm = data.index[0], data.index[0], data.index[-1]
        save_path = os.path.join(save_temp_dir, "{}.mp4".format(str(group_cnt)))
        _, video_name, _ = get_fn_ext(filepath=video_path)
        writer = cv2.VideoWriter(
            save_path,
            fourcc,
            video_meta_data["fps"],
            (video_meta_data["width"], video_meta_data["height"]),
        )
        cap = cv2.VideoCapture(video_path)
        cap.set(1, start_frm)
        color = colors[0]

        def __draw_individual_lines(animal_img_data: pd.DataFrame, img: np.array):
            color = colors[0]
            for cnt, (i, r) in enumerate(animal_img_data.iterrows()):
                if style_attr["Direction_color"] == "Random":
                    color = random.sample(colors[0], 1)[0]
                cv2.line(
                    img,
                    (int(r["Eye_x"]), int(r["Eye_y"])),
                    (int(r["Animal_2_bodypart_x"]), int(r["Animal_2_bodypart_y"])),
                    color,
                    style_attr["Direction_thickness"],
                )
                if style_attr["Highlight_endpoints"]:
                    cv2.circle(
                        img,
                        (int(r["Eye_x"]), int(r["Eye_y"])),
                        style_attr["Pose_circle_size"] + 2,
                        color,
                        style_attr["Pose_circle_size"],
                    )
                    cv2.circle(
                        img,
                        (int(r["Animal_2_bodypart_x"]), int(r["Animal_2_bodypart_y"])),
                        style_attr["Pose_circle_size"] + 1,
                        color,
                        style_attr["Pose_circle_size"],
                    )

            return img

        while current_frm < end_frm:
            ret, img = cap.read()
            try:
                if ret:
                    if style_attr["Show_pose"]:
                        bp_data = data.loc[current_frm]
                        for cnt, (animal_name, animal_bps) in enumerate(
                            bp_names.items()
                        ):
                            for bp in zip(animal_bps["X_bps"], animal_bps["Y_bps"]):
                                x_bp, y_bp = bp_data[bp[0]], bp_data[bp[1]]
                                cv2.circle(
                                    img,
                                    (int(x_bp), int(y_bp)),
                                    style_attr["Pose_circle_size"],
                                    bp_names[animal_name]["colors"][cnt],
                                    style_attr["Direction_thickness"],
                                )

                    if current_frm in list(directionality_data["Frame_#"].unique()):
                        img_data = directionality_data[
                            directionality_data["Frame_#"] == current_frm
                        ]
                        unique_animals = img_data["Animal_1"].unique()
                        for animal in unique_animals:
                            animal_img_data = img_data[
                                img_data["Animal_1"] == animal
                            ].reset_index(drop=True)
                            if style_attr["Polyfill"]:
                                convex_hull_arr = animal_img_data.loc[
                                    0, ["Eye_x", "Eye_y"]
                                ].values.reshape(-1, 2)
                                for animal_2 in animal_img_data["Animal_2"].unique():
                                    convex_hull_arr = np.vstack(
                                        (
                                            convex_hull_arr,
                                            animal_img_data[
                                                [
                                                    "Animal_2_bodypart_x",
                                                    "Animal_2_bodypart_y",
                                                ]
                                            ][
                                                animal_img_data["Animal_2"] == animal_2
                                            ].values,
                                        )
                                    ).astype("int")
                                    convex_hull_arr = np.unique(convex_hull_arr, axis=0)
                                    if convex_hull_arr.shape[0] >= 3:
                                        if style_attr["Direction_color"] == "Random":
                                            color = random.sample(colors[0], 1)[0]
                                        cv2.fillPoly(img, [convex_hull_arr], color)
                                    else:
                                        img = __draw_individual_lines(
                                            animal_img_data=animal_img_data, img=img
                                        )

                            else:
                                img = __draw_individual_lines(
                                    animal_img_data=animal_img_data, img=img
                                )

                    img = np.uint8(img)

                    current_frm += 1
                    writer.write(img)
                    print(
                        "Multi-processing video frame {} on core {}...".format(
                            str(current_frm), str(group_cnt)
                        )
                    )

                else:
                    cap.release()
                    writer.release()
                    break

            except IndexError:
                cap.release()
                writer.release()
                break

        return group_cnt

    @staticmethod
    def distance_plotter_mp(
        data: np.array,
        video_setting: bool,
        frame_setting: bool,
        video_name: str,
        video_save_dir: str,
        frame_folder_dir: str,
        style_attr: dict,
        line_attr: dict,
        fps: int,
    ):

        group = int(data[0][0])
        line_data = data[:, 2:]
        color_dict = get_color_dict()
        video_writer = None
        if video_setting:
            fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
            video_save_path = os.path.join(video_save_dir, "{}.mp4".format(str(group)))
            video_writer = cv2.VideoWriter(
                video_save_path,
                fourcc,
                fps,
                (style_attr["width"], style_attr["height"]),
            )

        for i in range(line_data.shape[0]):
            frame_id = int(data[i][1])
            for j in range(line_data.shape[1]):
                color = color_dict[line_attr[j][-1]][::-1]
                color = tuple(x / 255 for x in color)
                plt.plot(
                    line_data[0:i, j],
                    color=color,
                    linewidth=style_attr["line width"],
                    alpha=style_attr["opacity"],
                )

            x_ticks_locs = x_lbls = np.round(np.linspace(0, i, 5))
            x_lbls = np.round((x_lbls / fps), 1)
            plt.ylim(0, style_attr["max_y"])
            plt.xlabel("time (s)")
            plt.ylabel("distance (cm)")
            plt.xticks(
                x_ticks_locs,
                x_lbls,
                rotation="horizontal",
                fontsize=style_attr["font size"],
            )
            plt.yticks(
                style_attr["y_ticks_locs"],
                style_attr["y_ticks_lbls"],
                fontsize=style_attr["font size"],
            )
            plt.suptitle(
                "Animal distances", x=0.5, y=0.92, fontsize=style_attr["font size"] + 4
            )

            buffer_ = io.BytesIO()
            plt.savefig(buffer_, format="png")
            buffer_.seek(0)
            img = PIL.Image.open(buffer_)
            img = np.uint8(cv2.cvtColor(np.asarray(img), cv2.COLOR_RGB2BGR))
            buffer_.close()
            plt.close()

            img = cv2.resize(img, (style_attr["width"], style_attr["height"]))
            if video_setting:
                video_writer.write(np.uint8(img))
            if frame_setting:
                frm_name = os.path.join(frame_folder_dir, str(frame_id) + ".png")
                cv2.imwrite(frm_name, np.uint8(img))

            print(
                "Distance frame created: {}, Video: {}, Processing core: {}".format(
                    str(frame_id + 1), video_name, str(group + 1)
                )
            )

        return group

    @staticmethod
    def gantt_creator_mp(
        data: np.array,
        frame_setting: bool,
        video_setting: bool,
        video_save_dir: str,
        frame_folder_dir: str,
        bouts_df: pd.DataFrame,
        clf_names: list,
        colors: list,
        color_tuple: tuple,
        fps: int,
        rotation: int,
        font_size: int,
        width: int,
        height: int,
        video_name: str,
    ):

        group, frame_rng = data[0], data[1:]
        start_frm, end_frm, current_frm = frame_rng[0], frame_rng[-1], frame_rng[0]
        video_writer = None
        if video_setting:
            fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
            video_save_path = os.path.join(video_save_dir, "{}.mp4".format(str(group)))
            video_writer = cv2.VideoWriter(
                video_save_path, fourcc, fps, (width, height)
            )

        while current_frm < end_frm:
            fig, ax = plt.subplots()
            bout_rows = bouts_df.loc[bouts_df["End_frame"] <= current_frm]
            for i, event in enumerate(bout_rows.groupby("Event")):
                for x in clf_names:
                    if event[0] == x:
                        ix = clf_names.index(x)
                        data_event = event[1][["Start_time", "Bout_time"]]
                        ax.broken_barh(
                            data_event.values,
                            (color_tuple[ix], 3),
                            facecolors=colors[ix],
                        )

            x_ticks_locs = x_lbls = np.round(
                np.linspace(0, round((current_frm / fps), 3), 6)
            )
            ax.set_xticks(x_ticks_locs)
            ax.set_xticklabels(x_lbls)
            ax.set_ylim(0, color_tuple[len(clf_names)])
            ax.set_yticks(np.arange(5, 5 * len(clf_names) + 1, 5))
            ax.tick_params(axis="both", labelsize=font_size)
            ax.set_yticklabels(clf_names, rotation=rotation)
            ax.set_xlabel("Session (s)", fontsize=font_size)
            ax.yaxis.grid(True)
            canvas = FigureCanvas(fig)
            canvas.draw()
            mat = np.array(canvas.renderer._renderer)
            img = cv2.cvtColor(mat, cv2.COLOR_RGB2BGR)
            img = np.uint8(cv2.resize(img, (width, height)))
            if video_setting:
                video_writer.write(img)
            if frame_setting:
                frame_save_name = os.path.join(
                    frame_folder_dir, "{}.png".format(str(current_frm))
                )
                cv2.imwrite(frame_save_name, img)
            plt.close(fig)
            current_frm += 1

            print(
                "Gantt frame created: {}, Video: {}, Processing core: {}".format(
                    str(current_frm + 1), video_name, str(group + 1)
                )
            )

        if video_setting:
            video_writer.release()

        return group

    @staticmethod
    def roi_plotter_mp(
        data: pd.DataFrame,
        loc_dict: dict,
        scalers: dict,
        video_meta_data: dict,
        save_temp_directory: str,
        shape_meta_data: dict,
        video_shape_names: list,
        input_video_path: str,
        body_part_dict: dict,
        roi_analyzer_data: object,
        colors: list,
        style_attr: dict,
        animal_ids: list,
        threshold: float,
    ):

        def __insert_texts(shape_df):
            for animal_name in animal_ids:
                for _, shape in shape_df.iterrows():
                    shape_name, shape_color = shape["Name"], shape["Color BGR"]
                    cv2.putText(
                        border_img,
                        loc_dict[animal_name][shape_name]["timer_text"],
                        loc_dict[animal_name][shape_name]["timer_text_loc"],
                        font,
                        scalers["font_size"],
                        shape_color,
                        TextOptions.TEXT_THICKNESS.value,
                    )
                    cv2.putText(
                        border_img,
                        loc_dict[animal_name][shape_name]["entries_text"],
                        loc_dict[animal_name][shape_name]["entries_text_loc"],
                        font,
                        scalers["font_size"],
                        shape_color,
                        TextOptions.TEXT_THICKNESS.value,
                    )

            return border_img

        font = cv2.FONT_HERSHEY_TRIPLEX
        fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        group_cnt = int(data["group"].values[0])
        start_frm, current_frm, end_frm = data.index[0], data.index[0], data.index[-1]
        save_path = os.path.join(save_temp_directory, "{}.mp4".format(str(group_cnt)))
        writer = cv2.VideoWriter(
            save_path,
            fourcc,
            video_meta_data["fps"],
            (video_meta_data["width"] * 2, video_meta_data["height"]),
        )
        cap = cv2.VideoCapture(input_video_path)
        cap.set(1, start_frm)

        while current_frm < end_frm:
            ret, img = cap.read()
            border_img = cv2.copyMakeBorder(
                img,
                0,
                0,
                0,
                int(video_meta_data["width"]),
                borderType=cv2.BORDER_CONSTANT,
                value=[0, 0, 0],
            )
            border_img = __insert_texts(roi_analyzer_data.video_recs)
            border_img = __insert_texts(roi_analyzer_data.video_circs)
            border_img = __insert_texts(roi_analyzer_data.video_polys)

            for _, row in roi_analyzer_data.video_recs.iterrows():
                top_left_x, top_left_y, shape_name = (
                    row["topLeftX"],
                    row["topLeftY"],
                    row["Name"],
                )
                bottom_right_x, bottom_right_y = (
                    row["Bottom_right_X"],
                    row["Bottom_right_Y"],
                )
                thickness, color = row["Thickness"], row["Color BGR"]
                cv2.rectangle(
                    border_img,
                    (int(top_left_x), int(top_left_y)),
                    (int(bottom_right_x), int(bottom_right_y)),
                    color,
                    int(thickness),
                )

            for _, row in roi_analyzer_data.video_circs.iterrows():
                center_x, center_y, radius, shape_name = (
                    row["centerX"],
                    row["centerY"],
                    row["radius"],
                    row["Name"],
                )
                thickness, color = row["Thickness"], row["Color BGR"]
                cv2.circle(
                    border_img, (center_x, center_y), radius, color, int(thickness)
                )

            for _, row in roi_analyzer_data.video_polys.iterrows():
                vertices, shape_name = row["vertices"], row["Name"]
                thickness, color = row["Thickness"], row["Color BGR"]
                cv2.polylines(
                    border_img, [vertices], True, color, thickness=int(thickness)
                )

            for animal_cnt, animal_name in enumerate(animal_ids):
                if style_attr["Show_body_part"] or style_attr["Show_animal_name"]:
                    bp_data = data.loc[current_frm, body_part_dict[animal_name]].values
                    if threshold < bp_data[2]:
                        if style_attr["Show_body_part"]:
                            cv2.circle(
                                border_img,
                                (int(bp_data[0]), int(bp_data[1])),
                                scalers["circle_size"],
                                colors[animal_cnt],
                                -1,
                            )
                        if style_attr["Show_animal_name"]:
                            cv2.putText(
                                border_img,
                                animal_name,
                                (int(bp_data[0]), int(bp_data[1])),
                                font,
                                scalers["font_size"],
                                colors[animal_cnt],
                                TextOptions.TEXT_THICKNESS.value,
                            )

                for shape_name in video_shape_names:
                    timer = round(
                        data.loc[
                            current_frm,
                            "{}_{}_cum_sum_time".format(animal_name, shape_name),
                        ],
                        2,
                    )
                    entries = data.loc[
                        current_frm,
                        "{}_{}_cum_sum_entries".format(animal_name, shape_name),
                    ]
                    cv2.putText(
                        border_img,
                        str(timer),
                        loc_dict[animal_name][shape_name]["timer_data_loc"],
                        font,
                        scalers["font_size"],
                        shape_meta_data[shape_name]["Color BGR"],
                        TextOptions.TEXT_THICKNESS.value,
                    )
                    cv2.putText(
                        border_img,
                        str(entries),
                        loc_dict[animal_name][shape_name]["entries_data_loc"],
                        font,
                        scalers["font_size"],
                        shape_meta_data[shape_name]["Color BGR"],
                        TextOptions.TEXT_THICKNESS.value,
                    )

            writer.write(border_img)
            current_frm += 1
            print(
                "Multi-processing video frame {} on core {}...".format(
                    str(current_frm), str(group_cnt)
                )
            )

        cap.release()
        writer.release()

        return group_cnt

    @staticmethod
    def validation_video_mp(
        data: pd.DataFrame,
        bp_dict: dict,
        video_save_dir: str,
        settings: dict,
        video_path: str,
        video_meta_data: dict,
        gantt_setting: Optional[
            Literal[
                "Gantt chart: final frame only (slightly faster)", "Gantt chart: video"
            ]
        ],
        final_gantt: Optional[np.ndarray],
        clf_data: np.ndarray,
        clrs: List[List],
        clf_name: str,
        bouts_df: pd.DataFrame,
    ):

        dpi = plt.rcParams["figure.dpi"]

        def create_gantt(
            bouts_df: pd.DataFrame, clf_name: str, image_index: int, fps: int
        ):

            fig, ax = plt.subplots(
                figsize=(final_gantt.shape[1] / dpi, final_gantt.shape[0] / dpi)
            )
            matplotlib.font_manager._get_font.cache_clear()
            relRows = bouts_df.loc[bouts_df["End_frame"] <= image_index]

            for i, event in enumerate(relRows.groupby("Event")):
                data_event = event[1][["Start_time", "Bout_time"]]
                ax.broken_barh(data_event.values, (4, 4), facecolors="red")
            xLength = (round(image_index / fps)) + 1
            if xLength < 10:
                xLength = 10

            ax.set_xlim(0, xLength)
            ax.set_ylim([0, 12])
            ax.set_xlabel("Session (s)", fontsize=12)
            ax.set_ylabel(clf_name, fontsize=12)
            ax.set_title(f"{clf_name} GANTT CHART", fontsize=12)
            ax.set_yticks([])
            ax.yaxis.set_ticklabels([])
            ax.yaxis.grid(True)
            canvas = FigureCanvas(fig)
            canvas.draw()
            img = np.array(np.uint8(np.array(canvas.renderer._renderer)))[:, :, :3]
            plt.close(fig)
            return img

        fourcc, font = cv2.VideoWriter_fourcc(*"mp4v"), cv2.FONT_HERSHEY_COMPLEX
        cap = cv2.VideoCapture(video_path)
        group = data["group"].iloc[0]
        start_frm, current_frm, end_frm = data.index[0], data.index[0], data.index[-1]
        video_save_path = os.path.join(video_save_dir, "{}.mp4".format(str(group)))
        if gantt_setting is not None:
            writer = cv2.VideoWriter(
                video_save_path,
                fourcc,
                video_meta_data["fps"],
                (
                    int(video_meta_data["width"] + final_gantt.shape[1]),
                    int(video_meta_data["height"]),
                ),
            )
        else:
            writer = cv2.VideoWriter(
                video_save_path,
                fourcc,
                video_meta_data["fps"],
                (video_meta_data["width"], video_meta_data["height"]),
            )

        cap.set(1, start_frm)
        while current_frm < end_frm:
            clf_frm_cnt = np.sum(clf_data[0:current_frm])
            ret, img = cap.read()
            if settings["pose"]:
                for animal_cnt, (animal_name, animal_data) in enumerate(
                    bp_dict.items()
                ):
                    for bp_cnt, bp in enumerate(range(len(animal_data["X_bps"]))):
                        x_header, y_header = (
                            animal_data["X_bps"][bp],
                            animal_data["Y_bps"][bp],
                        )
                        animal_cords = tuple(
                            data.loc[current_frm, [x_header, y_header]]
                        )
                        cv2.circle(
                            img,
                            (int(animal_cords[0]), int(animal_cords[1])),
                            0,
                            clrs[animal_cnt][bp_cnt],
                            settings["styles"]["circle size"],
                        )
            if settings["animal_names"]:
                for animal_cnt, (animal_name, animal_data) in enumerate(
                    bp_dict.items()
                ):
                    x_header, y_header = (
                        animal_data["X_bps"][0],
                        animal_data["Y_bps"][0],
                    )
                    animal_cords = tuple(data.loc[current_frm, [x_header, y_header]])
                    cv2.putText(
                        img,
                        animal_name,
                        (int(animal_cords[0]), int(animal_cords[1])),
                        font,
                        settings["styles"]["font size"],
                        clrs[animal_cnt][0],
                        1,
                    )
            target_timer = round((1 / video_meta_data["fps"]) * clf_frm_cnt, 2)
            cv2.putText(
                img,
                "Timer",
                (10, settings["styles"]["space_scale"]),
                font,
                settings["styles"]["font size"],
                (0, 255, 0),
                2,
            )
            addSpacer = 2
            cv2.putText(
                img,
                (f"{clf_name} {target_timer}s"),
                (10, settings["styles"]["space_scale"] * addSpacer),
                font,
                settings["styles"]["font size"],
                (0, 0, 255),
                2,
            )
            addSpacer += 1
            cv2.putText(
                img,
                "Ensemble prediction",
                (10, settings["styles"]["space_scale"] * addSpacer),
                font,
                settings["styles"]["font size"],
                (0, 255, 0),
                2,
            )
            addSpacer += 2
            if clf_data[current_frm] == 1:
                cv2.putText(
                    img,
                    clf_name,
                    (10, +settings["styles"]["space_scale"] * addSpacer),
                    font,
                    settings["styles"]["font size"],
                    (2, 166, 249),
                    2,
                )
                addSpacer += 1
            if gantt_setting == "Gantt chart: final frame only (slightly faster)":
                img = np.concatenate((img, final_gantt), axis=1)
            elif gantt_setting == "Gantt chart: video":
                gantt_img = create_gantt(
                    bouts_df, clf_name, current_frm, video_meta_data["fps"]
                )
                img = np.concatenate((img, gantt_img), axis=1)

            writer.write(np.uint8(img))
            current_frm += 1
            print(
                "Multi-processing video frame {} on core {}...".format(
                    str(current_frm), str(group)
                )
            )

        cap.release()
        writer.release()

        return group

    @staticmethod
    def bbox_mp(
        frm_range: list,
        polygon_data: dict,
        animal_bp_dict: dict,
        data_df: Optional[pd.DataFrame],
        intersection_data_df: Optional[pd.DataFrame],
        roi_attributes: dict,
        video_path: str,
        key_points: bool,
        greyscale: bool,
    ):

        cap, current_frame = cv2.VideoCapture(video_path), frm_range[0]
        cap.set(1, frm_range[0])
        img_lst = []
        while current_frame < frm_range[-1]:
            ret, frame = cap.read()
            if ret:
                if key_points:
                    frm_data = data_df.iloc[current_frame]
                if greyscale:
                    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
                    frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)
                for animal_cnt, (animal, animal_data) in enumerate(
                    animal_bp_dict.items()
                ):
                    if key_points:
                        for bp_cnt, (x_col, y_col) in enumerate(
                            zip(animal_data["X_bps"], animal_data["Y_bps"])
                        ):
                            cv2.circle(
                                frame,
                                (frm_data[x_col], frm_data[y_col]),
                                0,
                                roi_attributes[animal]["bbox_clr"],
                                roi_attributes[animal]["keypoint_size"],
                            )
                    animal_polygon = np.array(
                        list(
                            polygon_data[animal][
                                current_frame
                            ].convex_hull.exterior.coords
                        )
                    ).astype(int)
                    if intersection_data_df is not None:
                        intersect = intersection_data_df.loc[
                            current_frame,
                            intersection_data_df.columns.str.startswith(animal),
                        ].sum()
                        if intersect > 0:
                            cv2.polylines(
                                frame,
                                [animal_polygon],
                                1,
                                roi_attributes[animal]["highlight_clr"],
                                roi_attributes[animal]["highlight_clr_thickness"],
                            )
                    cv2.polylines(
                        frame,
                        [animal_polygon],
                        1,
                        roi_attributes[animal]["bbox_clr"],
                        roi_attributes[animal]["bbox_thickness"],
                    )
                img_lst.append(frame)
                current_frame += 1
            else:
                print(
                    "SIMBA WARNING: SimBA tried to grab frame number {} from video {}, but could not find it. The video has {} frames.".format(
                        str(current_frame),
                        video_path,
                        str(cap.get(cv2.CAP_PROP_FRAME_COUNT)),
                    )
                )
        return img_lst

    @staticmethod
    def path_plot_mp(
        data: np.array,
        video_setting: bool,
        frame_setting: bool,
        video_save_dir: str,
        video_name: str,
        frame_folder_dir: str,
        style_attr: dict,
        print_animal_names: bool,
        animal_attr: dict,
        fps: int,
        video_info: pd.DataFrame,
        clf_attr: dict,
        input_style_attr: dict,
        video_path: Optional[Union[str, os.PathLike]] = None,
    ):

        group = int(data[0][0][0])
        color_dict = get_color_dict()
        if video_setting:
            fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
            video_save_path = os.path.join(video_save_dir, "{}.mp4".format(str(group)))
            video_writer = cv2.VideoWriter(
                video_save_path,
                fourcc,
                fps,
                (style_attr["width"], style_attr["height"]),
            )

        if input_style_attr is not None:
            if (type(input_style_attr["bg color"]) == dict) and (
                input_style_attr["bg color"]["type"]
            ) == "moving":
                check_file_exist_and_readable(file_path=video_path)
                video_cap = cv2.VideoCapture(video_path)

        for i in range(data.shape[0]):
            if input_style_attr is not None:
                if (type(input_style_attr["bg color"]) == dict) and (
                    input_style_attr["bg color"]["type"]
                ) == "moving":
                    style_attr["bg color"] = input_style_attr["bg color"]
            frame_id = int(data[i, -1, 1] + 1)
            frame_data = data[i, :, 2:].astype(int)
            frame_data = np.split(frame_data, len(list(animal_attr.keys())), axis=1)
            img = np.zeros(
                (
                    int(video_info["Resolution_height"].values[0]),
                    int(video_info["Resolution_width"].values[0]),
                    3,
                )
            )
            if (type(style_attr["bg color"]) == dict) and (
                style_attr["bg color"]["type"]
            ) == "moving":
                style_attr["bg color"] = read_frm_of_video(
                    video_path=video_cap,
                    opacity=style_attr["bg color"]["opacity"],
                    frame_index=frame_id,
                )

            img[:] = style_attr["bg color"]
            for animal_cnt, animal_data in enumerate(frame_data):
                animal_clr = style_attr["animal clrs"][animal_cnt]
                for j in range(animal_data.shape[0] - 1):
                    cv2.line(
                        img,
                        tuple(animal_data[j]),
                        tuple(animal_data[j + 1]),
                        animal_clr,
                        style_attr["line width"],
                    )
                cv2.circle(
                    img,
                    tuple(animal_data[-1]),
                    0,
                    animal_clr,
                    style_attr["circle size"],
                )
                if print_animal_names:
                    cv2.putText(
                        img,
                        style_attr["animal names"][animal_cnt],
                        tuple(animal_data[-1]),
                        cv2.FONT_HERSHEY_COMPLEX,
                        style_attr["font size"],
                        animal_clr,
                        style_attr["font thickness"],
                    )

            if clf_attr:
                for clf_cnt, clf_name in enumerate(clf_attr["data"].columns):
                    clf_size = int(clf_attr["attr"][clf_cnt][-1].split(": ")[-1])
                    clf_clr = color_dict[clf_attr["attr"][clf_cnt][1]]
                    clf_sliced = clf_attr["data"][clf_name].loc[0:frame_id]
                    clf_sliced_idx = list(clf_sliced[clf_sliced == 1].index)
                    locations = clf_attr["positions"][clf_sliced_idx, :]
                    for i in range(locations.shape[0]):
                        cv2.circle(
                            img,
                            (locations[i][0], locations[i][1]),
                            0,
                            clf_clr,
                            clf_size,
                        )

            img = cv2.resize(img, (style_attr["width"], style_attr["height"]))
            if video_setting:
                video_writer.write(np.uint8(img))
            if frame_setting:
                frm_name = os.path.join(frame_folder_dir, str(frame_id) + ".png")
                cv2.imwrite(frm_name, np.uint8(img))

            print(
                "Path frame created: {}, Video: {}, Processing core: {}".format(
                    str(frame_id + 1), video_name, str(group + 1)
                )
            )

        return group

    @staticmethod
    def probability_plot_mp(
        data: list,
        probability_lst: list,
        clf_name: str,
        video_setting: bool,
        frame_setting: bool,
        video_dir: str,
        frame_dir: str,
        highest_p: float,
        fps: int,
        style_attr: dict,
        video_name: str,
    ):

        group, data = data[0], data[1:]
        start_frm, end_frm, current_frm = data[0], data[-1], data[0]

        if video_setting:
            fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
            video_save_path = os.path.join(video_dir, "{}.mp4".format(str(group)))
            video_writer = cv2.VideoWriter(
                video_save_path,
                fourcc,
                fps,
                (style_attr["width"], style_attr["height"]),
            )

        while current_frm < end_frm:
            fig, ax = plt.subplots()
            current_lst = probability_lst[0 : current_frm + 1]
            ax.plot(
                current_lst,
                color=style_attr["color"],
                linewidth=style_attr["line width"],
            )
            ax.plot(
                current_frm,
                current_lst[-1],
                "o",
                markersize=style_attr["circle size"],
                color=style_attr["color"],
            )
            ax.set_ylim([0, highest_p])
            x_ticks_locs = x_lbls = np.linspace(0, current_frm, 5)
            x_lbls = np.round((x_lbls / fps), 1)
            ax.xaxis.set_ticks(x_ticks_locs)
            ax.set_xticklabels(x_lbls, fontsize=style_attr["font size"])
            ax.set_xlabel("Time (s)", fontsize=style_attr["font size"])
            ax.set_ylabel(
                "{} {}".format(clf_name, "probability"),
                fontsize=style_attr["font size"],
            )
            plt.suptitle(clf_name, x=0.5, y=0.92, fontsize=style_attr["font size"] + 4)
            canvas = FigureCanvas(fig)
            canvas.draw()
            mat = np.array(canvas.renderer._renderer)
            image = cv2.cvtColor(mat, cv2.COLOR_RGB2BGR)
            image = np.uint8(
                cv2.resize(image, (style_attr["width"], style_attr["height"]))
            )
            if video_setting:
                video_writer.write(image)
            if frame_setting:
                frame_save_name = os.path.join(
                    frame_dir, "{}.png".format(str(current_frm))
                )
                cv2.imwrite(frame_save_name, image)
            plt.close()
            current_frm += 1

            print(
                "Probability frame created: {}, Video: {}, Processing core: {}".format(
                    str(current_frm + 1), video_name, str(group + 1)
                )
            )

        return group

    def violin_plot(
        self,
        data: pd.DataFrame,
        x: str,
        y: str,
        save_path: Union[str, os.PathLike],
        font_rotation: Optional[int] = 45,
        font_size: Optional[int] = 10,
        img_size: Optional[tuple] = (13.7, 8.27),
        cut: Optional[int] = 0,
        scale: Optional[Literal["area", "count", "width"]] = "area",
    ):
        named_colors = get_named_colors()
        palette = {}
        for cnt, violin in enumerate(sorted(list(data[x].unique()))):
            palette[violin] = named_colors[cnt]
        plt.figure()
        order = data.groupby(by=[x])[y].median().sort_values().iloc[::-1].index
        figure_FSCTT = sns.violinplot(
            x=x, y=y, data=data, cut=cut, scale=scale, order=order, palette=palette
        )
        figure_FSCTT.set_xticklabels(
            figure_FSCTT.get_xticklabels(), rotation=font_rotation, size=font_size
        )
        figure_FSCTT.figure.set_size_inches(img_size)
        figure_FSCTT.figure.savefig(save_path, bbox_inches="tight")
        stdout_success(
            msg=f"Violin plot saved at {save_path}", source=self.__class__.__name__
        )

    @staticmethod
    @njit("(uint8[:,:,:],)")
    def rotate_img(img: np.ndarray):
        """
        Jitted helper to flip image 90 degrees.

        :example:
        >>> img = cv2.imread('/Users/simon/Desktop/sliding_line_length.png')
        >>> rotated_img = PlottingMixin().rotate_img(img)
        """
        rotated_image = np.transpose(img, (1, 0, 2))
        return np.ascontiguousarray(np.fliplr(rotated_image).astype(np.uint8))



__author__ = "Simon Nilsson"
import glob
import os
from typing import List, Union

import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from simba.mixins.config_reader import ConfigReader
from simba.utils.data import detect_bouts, plug_holes_shortest_bout
from simba.utils.errors import NotDirectoryError
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class AggBoundaryStatisticsCalculator(ConfigReader):
    """
    Compute aggregate boundary statistics

    :parameter str config_path: SimBA project config file in Configparser format
    :parameter List[str] measures: Aggregate statistics measurements. OPTIONS: 'DETAILED INTERACTIONS TABLE', 'INTERACTION TIME (s)', 'INTERACTION BOUT COUNT', 'INTERACTION BOUT MEAN (s)', 'INTERACTION BOUT MEDIAN (s)'
    :parameter int shortest_allowed_interaction: The shortest allowed animal-anchored ROI intersection in millisecond.

    Notes
    ----------
    `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/anchored_rois.md>`_.

    Examples
    ----------
    >>> boundary_stats_calculator = AggBoundaryStatisticsCalculator('MyProjectConfig', measures=['INTERACTION TIME (s)'], shortest_allowed_interaction=200)
    >>> boundary_stats_calculator.run()
    >>> boundary_stats_calculator.save()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        measures: List[
            Literal[
                "INTERACTION TIME (s)",
                "INTERACTION BOUT COUNT",
                "INTERACTION BOUT MEAN (s)",
                "INTERACTION BOUT MEDIAN (s)",
            ]
        ],
        shortest_allowed_interaction: int,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        self.measures, self.shortest_allowed_interaction_ms = (
            measures,
            shortest_allowed_interaction,
        )
        self.anchored_roi_path = os.path.join(
            self.project_path, "logs", "anchored_rois.pickle"
        )
        self.data_path = os.path.join(self.project_path, "csv", "anchored_roi_data")
        if not os.path.isdir(self.data_path):
            raise NotDirectoryError(
                msg=f"SIMBA ERROR: No anchored roi statistics found in {self.data_path}. Create data before analyzing aggregate statistics"
            )
        self.files_found = (
            glob.glob(self.data_path + "/*.pickle")
            + glob.glob(self.data_path + "/*.parquet")
            + glob.glob(self.data_path + "/*.csv")
        )

    def run(self):
        self.results = {}
        for file_cnt, file_path in enumerate(self.files_found):
            _, self.file_name, ext = get_fn_ext(file_path)
            print(f"Creating aggregate statistics for video {self.file_name}...")
            _, _, fps = self.read_video_info(video_name=self.file_name)
            data_df = read_df(file_path=file_path, file_type=ext[1:])
            if (self.shortest_allowed_interaction_ms / fps) > 0:
                for column in data_df.columns:
                    data_df = plug_holes_shortest_bout(
                        data_df=data_df,
                        clf_name=column,
                        fps=int(fps),
                        shortest_bout=self.shortest_allowed_interaction_ms,
                    )
            bouts_df = detect_bouts(
                data_df=data_df, target_lst=list(data_df.columns), fps=int(fps)
            )
            self.video_results, self.detailed_interactions_results = {}, {}
            if "INTERACTION TIME (s)" in self.measures:
                self.video_results["INTERACTION TIME (s)"] = (
                    bouts_df.groupby(by="Event")["Bout_time"].sum().to_dict()
                )
            if "INTERACTION BOUT COUNT" in self.measures:
                self.video_results["INTERACTION BOUT COUNT"] = (
                    bouts_df.groupby(by="Event")["Bout_time"].count().to_dict()
                )
            if "INTERACTION BOUT TIME MEAN (s)" in self.measures:
                self.video_results["INTERACTION BOUT MEAN (s)"] = (
                    bouts_df.groupby(by="Event")["Bout_time"].mean().to_dict()
                )
            if "INTERACTION BOUT TIME MEDIAN (s)" in self.measures:
                self.video_results["INTERACTION BOUT MEDIAN (s)"] = (
                    bouts_df.groupby(by="Event")["Bout_time"].median().to_dict()
                )
            if "DETAILED INTERACTIONS TABLE" in self.measures:
                self.create_detailed_interactions_table(df=bouts_df)
            self.results[self.file_name] = self.video_results

    def save(self):
        self.timer.stop_timer()
        save_path = os.path.join(
            self.project_path,
            "logs",
            "aggregate_statistics_anchored_rois_{}.csv".format(self.datetime),
        )
        out_df = pd.DataFrame(
            columns=[
                "VIDEO",
                "ANIMAL 1",
                "ANIMAL 2",
                "ANIMAL 2 KEYPOINT",
                "MEASUREMENT",
                "VALUE",
            ]
        )
        if len(self.results.keys()) > 0:
            for video, video_data in self.results.items():
                for measurement, measurement_data in video_data.items():
                    for (
                        animal_interaction,
                        animal_interaction_value,
                    ) in measurement_data.items():
                        animal_names = animal_interaction.split(":")
                        if len(animal_names) == 2:
                            animal_names.append("None")
                        out_df.loc[len(out_df)] = [
                            video,
                            animal_names[0],
                            animal_names[1],
                            animal_names[2],
                            measurement,
                            animal_interaction_value,
                        ]
            out_df["VALUE"] = out_df["VALUE"].round(4)
            out_df = out_df.sort_values(by=["VIDEO", "MEASUREMENT"]).set_index("VIDEO")
            out_df.to_csv(save_path)
            stdout_success(
                msg=f"Aggregate animal-anchored ROI statistics saved at {save_path}",
                elapsed_time=self.timer.elapsed_time_str,
            )
        if len(self.detailed_interactions_results.keys()) > 0:
            save_path = os.path.join(
                self.project_path,
                "logs",
                "detailed_aggregate_statistics_anchored_rois_{}.csv".format(
                    self.datetime
                ),
            )
            out_df = pd.concat(
                self.detailed_interactions_results.values(), ignore_index=True
            )
            out_df = out_df.sort_values(by=["VIDEO"]).set_index("VIDEO")
            out_df.to_csv(save_path)
            stdout_success(
                msg=f"Detailed Aggregate animal-anchored ROI statistics saved at {save_path}",
                elapsed_time=self.timer.elapsed_time_str,
            )

    def create_detailed_interactions_table(self, df: pd.DataFrame):
        df = df.rename(
            columns={
                "Start_time": "START TIME (s)",
                "End Time": "END TIME (s)",
                "Start_frame": "START FRAME",
                "End_frame": "END FRAME",
                "Bout_time": "BOUT TIME (s)",
            }
        )
        df["ROI 1"], df["ROI 2"], df["KEY-POINT"] = df["Event"].str.split(":", 2).str
        df = df.drop(["Event"], axis=1)
        df["VIDEO"] = self.file_name
        df["BOUT FRAMES"] = (df["END FRAME"] + 1) - df["START FRAME"]
        df = df[
            [
                "VIDEO",
                "ROI 1",
                "ROI 2",
                "KEY-POINT",
                "START TIME (s)",
                "END TIME (s)",
                "START FRAME",
                "END FRAME",
                "BOUT FRAMES",
                "BOUT TIME (s)",
            ]
        ]
        self.detailed_interactions_results[self.file_name] = df


# boundary_stats_calculator = AggBoundaryStatisticsCalculator('/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/project_config.ini',
#                                                             measures=['INTERACTION TIME (s)', 'DETAILED INTERACTIONS TABLE'], shortest_allowed_interaction=0)
# boundary_stats_calculator.run()
# boundary_stats_calculator.save()



import cv2
import numpy as np


def zoom_in(self, zoom_val):
    self.not_done = True

    def get_x_y_callback(event, x, y, flags, param):
        if event == 1:
            self.click_loc = (int(x), int(y))
            self.not_done = False

    def initiate_zoom_in_callback():
        while True:
            cv2.setMouseCallback("Define shape", get_x_y_callback)
            cv2.waitKey(20)
            if self.not_done == False:
                break

    def get_new_image_spec():
        px_avail_right, px_avail_bottom = (
            self.frame_width - self.click_loc[0],
            self.frame_height - self.click_loc[1],
        )
        px_avail_left, px_avail_top = (
            self.frame_width - px_avail_right,
            self.frame_height - px_avail_bottom,
        )

        if (px_avail_left < (self.new_img_size[0] / 2)) or (
            px_avail_right < (self.new_img_size[0] / 2)
        ):
            top_left_x = 0
            bottom_right_x = self.new_img_size[0]
        else:
            top_left_x = int(self.click_loc[0] - (self.new_img_size[0] / 2))
            bottom_right_x = int(self.click_loc[0] + (self.new_img_size[0] / 2))
        if (px_avail_top < (self.new_img_size[1] / 2)) or (
            px_avail_bottom < (self.new_img_size[1] / 2)
        ):
            top_left_y = 0
            bottom_right_y = self.new_img_size[1]
        else:
            top_left_y = int(self.click_loc[1] - (self.new_img_size[1] / 2))
            bottom_right_y = int(self.click_loc[1] + (self.new_img_size[1] / 2))

        self.working_frame = self.working_frame[
            top_left_y:bottom_right_y, top_left_x:bottom_right_x
        ]
        self.working_frame = cv2.resize(
            self.working_frame,
            (self.frame_width, self.frame_height),
            interpolation=cv2.INTER_AREA,
        )
        self.insert_all_ROIs_into_image(show_zoomed_img=True)

    self.zoom_pct = int(zoom_val) / 100
    self.current_zoom = self.current_zoom + zoom_val
    self.new_img_size = (
        int(self.frame_width - (self.frame_width * self.zoom_pct)),
        int(self.frame_height - (self.frame_height * self.zoom_pct)),
    )
    initiate_zoom_in_callback()
    get_new_image_spec()


# self.interact_method = interact_method
# print(self.interact_method)
#
# def zoom_callback(event, x, y, flags, param):
#     if event == 1:
#         self.zoom_center_X, self.zoom_center_Y = int(x), int(y)
#         self.new_cords_dict = {}
#         if self.interact_method is 'zoom_in':
#             self.top_left_x_new = int(self.zoom_center_X - (self.new_size[0] / 2))
#             if self.top_left_x_new < 0: self.top_left_x_new = 0
#             self.top_left_y_new = int(self.zoom_center_Y - (self.new_size[1] / 2))
#             if self.top_left_y_new < 0: self.top_left_y_new = 0
#             self.bottom_right_x_new = int(self.zoom_center_X + (self.new_size[0] / 2))
#             self.bottom_right_y_new = int(self.zoom_center_Y + (self.new_size[1] / 2))
#             self.cropped_img = self.zoomed_img[self.top_left_y_new:self.bottom_right_y_new, self.top_left_x_new:self.bottom_right_x_new]
#             self.cropped_img = cv2.resize(self.cropped_img, (self.frame_width, self.frame_height), interpolation = cv2.INTER_AREA)
#             self.zoomed_img = self.cropped_img
#             cv2.imshow('Define shape', self.cropped_img)
#
# if self.interact_method == 'zoom_in' or 'zoom_out':
#     self.zoom_pct = int(pct) / 100
#     if interact_method == 'zoom_in':
#         self.new_size = (self.frame_width - (self.frame_width * self.zoom_pct), self.frame_height - (self.frame_height * self.zoom_pct))
#         cv2.setMouseCallback('Define shape', zoom_callback)
#



import os
from tkinter import *

import simba
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.utils.enums import Paths


class AboutSimBAPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(self, size=(896, 507), title="ABOUT SIMBA")
        canvas = Canvas(self.main_frm, width=896, height=507, bg="black")
        canvas.pack()
        img_path = os.path.join(os.path.dirname(simba.__file__), Paths.ABOUT_ME.value)
        img = PhotoImage(file=os.path.join(img_path))
        canvas.create_image(0, 0, image=img, anchor=NW)
        canvas.image = img



__author__ = "Simon Nilsson"

import os
from subprocess import PIPE, Popen
from tkinter import *
from tkinter import filedialog
from typing import Dict, Optional, Union

import cv2
import pandas as pd
from PIL import Image, ImageTk
from tabulate import tabulate

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import simba
from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_int, check_that_column_exist)
from simba.utils.enums import Options, TagNames
from simba.utils.errors import FrameRangeError
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (get_all_clf_names, get_fn_ext,
                                    get_video_meta_data, read_config_entry,
                                    read_df, write_df)


class LabellingInterface(ConfigReader):
    """
    Launch ``standard`` or ``pseudo``-labelling (annotation) GUI interface in SimBA.

    .. note::
       Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/label_behavior.md>`__.

    .. image:: _static/img/annotator.png
       :width: 500
       :align: center


    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str file_path: Path to video that is to be annotated
    :parameter str setting: String representing annotation method. OPTIONS: ``from_scratch`` or ``pseudo``
    :parameter dict threshold_dict: If setting ``pseudo``, threshold_dict dict contains the machine probability thresholds, with the classifier
        names as keys and the classification probabilities as values, e.g. {'Attack': 0.40, 'Sniffing': 0.7).
    :parameter bool continuing: If True, continouing previously started annotation session.

    Examples
    ----------
    >>> select_labelling_video(config_path='MyConfigPath', threshold_dict={'Attack': 0.4}, file_path='MyVideoFilePath', setting='pseudo', continuing=False)
    """

    def __init__(
        self,
        config_path: str,
        file_path: str,
        threshold_dict: Optional[Dict[str, float]] = None,
        setting: Literal["from_scratch", "pseudo"] = "pseudo",
        continuing: bool = False,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        self.padding, self.file_path = 5, file_path
        self.frm_no, self.threshold_dict = 0, threshold_dict
        self.setting = setting
        self.play_video_script_path = os.path.join(
            os.path.dirname(simba.__file__), "labelling/play_annotation_video.py"
        )
        _, self.video_name, _ = get_fn_ext(filepath=file_path)
        self.features_extracted_file_path = os.path.join(
            self.features_dir, self.video_name + "." + self.file_type
        )
        self.targets_inserted_file_path = os.path.join(
            self.targets_folder, self.video_name + "." + self.file_type
        )
        self.machine_results_file_path = os.path.join(
            self.machine_results_dir, self.video_name + "." + self.file_type
        )
        self.video_path = file_path
        self.cap = cv2.VideoCapture(self.video_path)
        self.video_meta_data = get_video_meta_data(video_path=self.video_path)
        self.frame_lst = list(range(0, self.video_meta_data["frame_count"]))
        self.max_frm_no = max(self.frame_lst)
        self.target_lst = get_all_clf_names(config=self.config, target_cnt=self.clf_cnt)
        self.max_frm_size = 1080, 650
        self.main_window = Toplevel()
        if continuing:
            check_file_exist_and_readable(file_path=self.targets_inserted_file_path)
            check_file_exist_and_readable(file_path=self.features_extracted_file_path)
            self.data_df = read_df(self.targets_inserted_file_path, self.file_type)
            self.data_df_features = read_df(
                self.features_extracted_file_path, self.file_type
            )
            missing_idx = self.data_df_features.index.difference(self.data_df.index)
            if len(missing_idx) > 0:
                self.data_df_features = self.data_df_features.iloc[
                    self.data_df_features.index.difference(self.data_df.index)
                ]
                self.data_df = pd.concat(
                    [self.data_df.astype(int), self.data_df_features], axis=0
                ).sort_index()
            self.main_window.title(
                "SIMBA ANNOTATION INTERFACE (CONTINUING ANNOTATIONS) - {}".format(
                    self.video_name
                )
            )
            self.frm_no = read_config_entry(
                self.config,
                "Last saved frames",
                self.video_name,
                data_type="int",
                default_value=0,
            )

        else:
            if setting == "from_scratch":
                check_file_exist_and_readable(
                    file_path=self.features_extracted_file_path
                )
                self.data_df = read_df(
                    self.features_extracted_file_path, self.file_type
                )
                self.main_window.title(
                    "SIMBA ANNOTATION INTERFACE (ANNOTATING FROM SCRATCH) - {}".format(
                        self.video_name
                    )
                )
                for target in self.target_lst:
                    self.data_df[target] = 0
            elif setting == "pseudo":
                check_file_exist_and_readable(file_path=self.machine_results_file_path)
                self.data_df = read_df(self.machine_results_file_path, self.file_type)
                for target in self.target_lst:
                    self.data_df.loc[
                        self.data_df["Probability_{}".format(target)]
                        > self.threshold_dict[target],
                        target,
                    ] = 1
                    self.data_df.loc[
                        self.data_df["Probability_{}".format(target)]
                        <= self.threshold_dict[target],
                        target,
                    ] = 0
                self.main_window.title(
                    "SIMBA ANNOTATION INTERFACE (PSEUDO-LABELLING) - {}".format(
                        self.video_name
                    )
                )
        self.data_df_targets = self.data_df
        for target in self.target_lst:
            check_that_column_exist(
                df=self.data_df_targets, column_name=target, file_name=file_path
            )
        self.data_df_targets = self.data_df_targets[self.target_lst]
        self.folder = Frame(self.main_window)
        self.buttons_frm = Frame(self.main_window, bd=2, width=700, height=300)
        self.current_frm_n = IntVar(self.main_window, value=self.frm_no)
        self.change_frm_box = Entry(
            self.buttons_frm, width=7, textvariable=self.current_frm_n
        )
        self.frame_number_lbl = Label(self.buttons_frm, text="Frame number")
        self.forward_btn = Button(
            self.buttons_frm,
            text=">",
            command=lambda: self.__advance_frame(
                new_frm_number=int(self.current_frm_n.get() + 1)
            ),
        )
        self.backward_btn = Button(
            self.buttons_frm,
            text="<",
            command=lambda: self.__advance_frame(
                new_frm_number=int(self.current_frm_n.get() - 1)
            ),
        )
        self.forward_max_btn = Button(
            self.buttons_frm,
            text=">>",
            command=lambda: self.__advance_frame(len(self.frame_lst) - 1),
        )
        self.backward_max_btn = Button(
            self.buttons_frm, text="<<", command=lambda: self.__advance_frame(0)
        )
        self.select_frm_btn = Button(
            self.buttons_frm,
            text="Jump to selected frame",
            command=lambda: self.__advance_frame(
                new_frm_number=int(self.change_frm_box.get())
            ),
        )
        self.jump_frame = Frame(self.main_window)
        self.jump = Label(self.jump_frame, text="Jump Size:")
        self.jump_size = Scale(
            self.jump_frame, from_=0, to=100, orient=HORIZONTAL, length=200
        )
        self.jump_size.set(0)
        self.jump_back = Button(
            self.jump_frame,
            text="<<",
            command=lambda: self.__advance_frame(
                int(self.change_frm_box.get()) - self.jump_size.get()
            ),
        )
        self.jump_forward = Button(
            self.jump_frame,
            text=">>",
            command=lambda: self.__advance_frame(
                int(self.change_frm_box.get()) + self.jump_size.get()
            ),
        )

        self.folder.grid(row=0, column=1, sticky=N)
        self.buttons_frm.grid(row=1, column=0)
        self.change_frm_box.grid(row=0, column=1)
        self.forward_btn.grid(row=1, column=3, sticky=E, padx=self.padding)
        self.backward_btn.grid(row=1, column=1, sticky=W, padx=self.padding)
        self.change_frm_box.grid(row=1, column=1)
        self.forward_max_btn.grid(row=1, column=4, sticky=W, padx=self.padding)
        self.backward_max_btn.grid(row=1, column=0, sticky=W, padx=self.padding)
        self.select_frm_btn.grid(row=2, column=1, sticky=N)
        self.jump_frame.grid(row=2, column=0)
        self.jump.grid(row=0, column=0, sticky=W)
        self.jump_size.grid(row=0, column=1, sticky=W)
        self.jump_back.grid(row=0, column=2, sticky=E)
        self.jump_forward.grid(row=0, column=3, sticky=W)

        self.check_frame = Frame(self.main_window, bd=2, width=300, height=500)
        self.check_frame.grid(row=0, column=1)
        self.check_behavior_lbl = Label(self.check_frame, text="Check Behavior:")
        self.check_behavior_lbl.config(font=("Calibri", 16))
        self.check_behavior_lbl.grid(sticky=N)

        self.checkboxes = {}
        for target_cnt, target in enumerate(self.target_lst):
            self.checkboxes[target] = {}
            self.checkboxes[target]["name"] = target
            self.checkboxes[target]["var"] = IntVar(
                value=self.data_df_targets[target].iloc[self.current_frm_n.get()]
            )
            self.checkboxes[target]["cb"] = Checkbutton(
                self.check_frame,
                text=target,
                variable=self.checkboxes[target]["var"],
                command=lambda k=self.checkboxes[target][
                    "name"
                ]: self.save_behavior_in_frm(
                    frame_number=int(self.current_frm_n.get()), target=k
                ),
            )
            self.checkboxes[target]["cb"].grid(row=target_cnt + 1, sticky=W)

        self.range_on = IntVar(value=0)
        self.range_frames = Frame(self.main_window)
        self.range_frames.grid(row=1, column=1, sticky=S)
        self.select_range = Checkbutton(
            self.range_frames, text="Frame range", variable=self.range_on
        )
        self.select_range.grid(row=0, column=0, sticky=W)
        self.first_frame = Entry(self.range_frames, width=7)
        self.first_frame.grid(row=0, column=1, sticky=E)
        self.to_label = Label(self.range_frames, text=" to ")
        self.to_label.grid(row=0, column=2, sticky=E)
        self.last_frame = Entry(self.range_frames, width=7)
        self.last_frame.grid(row=0, column=3, sticky=E)

        save = Button(
            self.main_window,
            text="Save Range",
            command=lambda: self.__save_behavior_in_range(
                self.first_frame.get(), self.last_frame.get()
            ),
        )
        save.grid(row=2, column=1, sticky=N)

        self.generate = Button(
            self.main_window,
            text="Save Annotations",
            command=lambda: self.__save_results(),
            fg="blue",
        )
        self.generate.config(font=("Calibri", 16))
        self.generate.grid(row=10, column=1, sticky=N)

        self.video_player_frm = Frame(self.main_window, width=100, height=100)
        self.video_player_frm.grid(row=0, column=2, sticky=N)
        self.play_video_btn = Button(
            self.video_player_frm, text="Open Video", command=self.__play_video
        )
        self.play_video_btn.grid(sticky=N, pady=10)
        self.video_key_lbls = Label(
            self.video_player_frm,
            text="\n\n  Keyboard shortcuts for video navigation: \n p = Pause/Play"
            "\n\n After pressing pause:"
            "\n o = +2 frames \n e = +10 frames \n w = +1 second"
            "\n\n t = -2 frames \n s = -10 frames \n x = -1 second"
            "\n\n q = Close video window \n\n",
        )
        self.video_key_lbls.grid(sticky=W)
        self.update_img_from_video = Button(
            self.video_player_frm,
            text="Show current video frame",
            command=self.__update_frame_from_video,
        )
        self.update_img_from_video.grid(sticky=N)
        self.__bind_shortcut_keys()
        self.key_presses_lbl = Label(
            self.video_player_frm,
            text="\n\n Keyboard shortcuts for frame navigation: \n Right Arrow = +1 frame"
            "\n Left Arrow = -1 frame"
            "\n Ctrl + s = Save annotations file"
            "\n Ctrl + a = +1 frame and keep choices"
            "\n Ctrl + p = Show annotation statistics"
            "\n Ctrl + l = Last frame"
            "\n Ctrl + o = First frame",
        )
        self.key_presses_lbl.grid(sticky=S)
        self.__advance_frame(new_frm_number=self.frm_no)
        self.main_window.mainloop()

    def __bind_shortcut_keys(self):
        self.main_window.bind("<Control-s>", lambda x: self.__save_results())
        self.main_window.bind(
            "<Control-a>",
            lambda x: self.__advance_frame(
                new_frm_number=int(self.current_frm_n.get() + 1),
                save_and_keep_checks=True,
            ),
        )
        self.main_window.bind(
            "<Control-p>", lambda x: self.print_annotation_statistics()
        )
        self.main_window.bind(
            "<Right>",
            lambda x: self.__advance_frame(
                new_frm_number=int(self.current_frm_n.get() + 1)
            ),
        )
        self.main_window.bind(
            "<Left>",
            lambda x: self.__advance_frame(
                new_frm_number=int(self.current_frm_n.get() - 1)
            ),
        )
        self.main_window.bind(
            "<Control-l>",
            lambda x: self.__advance_frame(new_frm_number=self.max_frm_no),
        )
        self.main_window.bind("<Control-o>", lambda x: self.__advance_frame(0))

    def print_annotation_statistics(self):
        table_view = [
            ["Video name", self.video_name],
            ["Video frames", self.video_meta_data["frame_count"]],
        ]
        for target in self.target_lst:
            present = len(self.data_df_targets[self.data_df_targets[target] == 1])
            absent = len(self.data_df_targets[self.data_df_targets[target] == 0])
            table_view.append([target + " present labels", present])
            table_view.append([target + " absent labels", absent])
            table_view.append(
                [target + " % present", present / self.video_meta_data["frame_count"]]
            )
            table_view.append(
                [target + " % absent", absent / self.video_meta_data["frame_count"]]
            )
        headers = ["VARIABLE", "VALUE"]
        print(tabulate(table_view, headers, tablefmt="github"))

    def __play_video(self):
        p = Popen(
            "python {}".format(self.play_video_script_path),
            stdin=PIPE,
            stdout=PIPE,
            shell=True,
        )
        main_project_dir = os.path.dirname(self.config_path)
        p.stdin.write(bytes(self.video_path, "utf-8"))
        p.stdin.close()
        temp_file = os.path.join(main_project_dir, "subprocess.txt")
        with open(temp_file, "w") as text_file:
            text_file.write(str(p.pid))

    def __update_frame_from_video(self):
        f = open(
            os.path.join(os.path.dirname(self.config_path), "labelling_info.txt"), "r+"
        )
        os.fsync(f.fileno())
        vid_frame_no = int(f.readline())
        self.__advance_frame(new_frm_number=vid_frame_no)
        f.close()

    def __read_frm(self, frm_number=None):
        self.cap.set(1, frm_number)
        _, self.current_frm_npy = self.cap.read()
        self.current_frm_npy = cv2.cvtColor(self.current_frm_npy, cv2.COLOR_RGB2BGR)
        self.current_frm_pil = Image.fromarray(self.current_frm_npy)
        self.current_frm_pil.thumbnail(self.max_frm_size, Image.LANCZOS)
        self.current_frm_pil = ImageTk.PhotoImage(
            master=self.main_window, image=self.current_frm_pil
        )
        self.video_frame = Label(self.main_window, image=self.current_frm_pil)
        self.video_frame.image = self.current_frm_pil
        self.video_frame.grid(row=0, column=0)

    def __advance_frame(self, new_frm_number: int, save_and_keep_checks=False):
        if new_frm_number > self.max_frm_no:
            print(
                "FRAME {} CANNOT BE SHOWN - YOU ARE VIEWING THE FINAL FRAME OF THE VIDEO (FRAME NUMBER {})".format(
                    str(new_frm_number), str(self.max_frm_no)
                )
            )
            self.current_frm_n = IntVar(value=self.max_frm_no)
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, self.current_frm_n.get())
        elif new_frm_number < 0:
            print(
                "FRAME {} CANNOT BE SHOWN - YOU ARE VIEWING THE FIRST FRAME OF THE VIDEO (FRAME NUMBER {})".format(
                    str(new_frm_number), str(self.max_frm_no)
                )
            )
            self.current_frm_n = IntVar(value=0)
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, self.current_frm_n.get())
        else:
            self.__create_print_statements()
            self.current_frm_n = IntVar(value=new_frm_number)
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, self.current_frm_n.get())
            if not save_and_keep_checks:
                for target in self.target_lst:
                    self.checkboxes[target]["var"].set(
                        bool(
                            self.data_df_targets[target].loc[
                                int(self.current_frm_n.get())
                            ]
                        )
                    )
            else:
                for target in self.target_lst:
                    self.checkboxes[target]["var"].set(
                        self.data_df_targets[target].loc[
                            int(self.current_frm_n.get() - 1)
                        ]
                    )
                    self.save_behavior_in_frm(target=target)
            self.__read_frm(frm_number=int(self.current_frm_n.get()))

    def __save_behavior_in_range(self, start_frm=None, end_frm=None):
        if not self.range_on.get():
            raise FrameRangeError(
                "SAVE RANGE ERROR: TO SAVE RANGE OF FRAMES, TICK THE `Frame range` checkbox before clicking `Save Range`",
                source=self.__class__.__name__,
            )
        else:
            check_int(
                "START FRAME", int(start_frm), max_value=self.max_frm_no, min_value=0
            )
            check_int("END FRAME", int(end_frm), max_value=self.max_frm_no, min_value=0)
            for frm_no in range(int(start_frm), int(end_frm) + 1):
                for target in self.target_lst:
                    self.data_df_targets[target].loc[frm_no] = self.checkboxes[target][
                        "var"
                    ].get()
            self.__read_frm(frm_number=int(end_frm))
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, end_frm)
        self.__create_print_statements(
            frame_range=True, start_frame=start_frm, end_frame=end_frm
        )

    def save_behavior_in_frm(self, frame_number=None, target=None):
        self.data_df_targets[target].loc[int(self.current_frm_n.get())] = int(
            self.checkboxes[target]["var"].get()
        )

    def __save_results(self):
        self.save_df = read_df(self.features_extracted_file_path, self.file_type)
        self.save_df = pd.concat([self.save_df, self.data_df_targets], axis=1)
        try:
            write_df(self.save_df, self.file_type, self.targets_inserted_file_path)
        except Exception as e:
            print(
                e,
                f"SIMBA ERROR: File for video {get_fn_ext(self.features_extracted_file_path)[1]} could not be saved.",
            )
            raise FileExistsError
        stdout_success(
            msg=f"SAVED: Annotation file for video {self.video_name} saved within the project_folder/csv/targets_inserted directory.",
            source=self.__class__.__name__,
        )
        if not self.config.has_section("Last saved frames"):
            self.config.add_section("Last saved frames")
        self.config.set(
            "Last saved frames", str(self.video_name), str(self.current_frm_n.get())
        )
        with open(self.config_path, "w") as configfile:
            self.config.write(configfile)

    def __create_print_statements(
        self, frame_range: bool = None, start_frame: int = None, end_frame: int = None
    ):
        print("USER FRAME SELECTION(S):")
        if not frame_range:
            for target in self.target_lst:
                target_present_choice = self.checkboxes[target]["var"].get()
                if target_present_choice == 0:
                    print(
                        "{} ABSENT IN FRAME {}".format(target, self.current_frm_n.get())
                    )
                if target_present_choice == 1:
                    print(
                        "{} PRESENT IN FRAME {}".format(
                            target, self.current_frm_n.get()
                        )
                    )

        if frame_range:
            for target in self.target_lst:
                target_present_choice = self.checkboxes[target]["var"].get()
                if target_present_choice == 1:
                    print(
                        "{} PRESENT IN FRAMES {} to {}".format(
                            target, str(start_frame), str(end_frame)
                        )
                    )
                elif target_present_choice == 0:
                    print(
                        "{} ABSENT IN FRAMES {} to {}".format(
                            target, str(start_frame), str(end_frame)
                        )
                    )


def select_labelling_video(
    config_path: Union[str, os.PathLike],
    threshold_dict: Optional[Dict[str, float]] = None,
    setting: str = None,
    continuing: bool = None,
    video_file_path=Union[str, os.PathLike],
):
    if setting is not "pseudo":
        video_file_path = filedialog.askopenfilename(
            filetypes=[("Video files", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)]
        )
    else:
        threshold_dict_values = {}
        for k, v in threshold_dict.items():
            check_float(name=k, value=float(v.entry_get), min_value=0.0, max_value=1.0)
            threshold_dict_values[k] = float(v.entry_get)
        threshold_dict = threshold_dict_values

    check_file_exist_and_readable(video_file_path)
    video_meta = get_video_meta_data(video_file_path)
    _, video_name, _ = get_fn_ext(video_file_path)

    print(f"ANNOTATING VIDEO {video_name} \n VIDEO INFO: {video_meta}")
    _ = LabellingInterface(
        config_path=config_path,
        file_path=video_file_path,
        threshold_dict=threshold_dict,
        setting=setting,
        continuing=continuing,
    )


# test = select_labelling_video(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                               threshold_dict={'Attack': 0.4},
#                               setting='from_scratch',
#                               continuing=True)



__author__ = "Simon Nilsson"

import itertools
import os
from typing import List, Union

import cv2
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.errors import NoROIDataError, NoSpecifiedOutputError
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, get_video_meta_data, read_df


class CueLightVisualizer(ConfigReader):
    """

    Visualize SimBA computed cue-light ON and OFF states and the aggregate statistics of ON and OFF
    states.

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter List[str] cue_light_names: Names of cue lights, as defined in the SimBA ROI interface.
    :parameter str video_path: Path to video which user wants to create visualizations of cue light states and aggregate statistics for.
    :parameter bool frame_setting: If True, creates individual frames in png format.
    :parameter bool video_setting: If True, creates compressed videos in mp4 format.

    .. notes:
       `Cue light tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/cue_light_tutorial.md>`__.

    Examples
    ----------
    >>> cue_light_visualizer = CueLightVisualizer(config_path='SimBAConfig', cue_light_names=['Cue_light'], video_path='VideoPath', video_setting=True, frame_setting=False)
    >>> cue_light_visualizer.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        cue_light_names: List[str],
        video_path: str,
        frame_setting: bool,
        video_setting: bool,
    ):
        ConfigReader.__init__(self, config_path=config_path)

        if (not frame_setting) and (not video_setting):
            raise NoSpecifiedOutputError(
                msg="SIMBA ERROR: Please choose to select either videos, frames, or both frames and videos."
            )
        self.video_setting, self.frame_setting = video_setting, frame_setting
        self.in_dir = os.path.join(self.project_path, "csv", "cue_lights")
        self.cue_light_names, self.video_path = cue_light_names, video_path
        _, self.video_name, _ = get_fn_ext(video_path)
        self.video_meta_data = get_video_meta_data(self.video_path)
        self.logs_path, self.video_dir = os.path.join(
            self.project_path, "logs"
        ), os.path.join(self.project_path, "videos")
        self.data_file_path = os.path.join(
            self.in_dir, self.video_name + "." + self.file_type
        )
        check_file_exist_and_readable(self.data_file_path)
        self.data_df = read_df(self.data_file_path, self.file_type)
        self.output_folder = os.path.join(
            self.project_path, "frames", "output", "cue_lights"
        )
        if not os.path.exists(self.output_folder):
            os.makedirs(self.output_folder)
        self.video_settings, pix_per_mm, self.fps = self.read_video_info(
            video_name=self.video_name
        )
        self.space_scale, radius_scale, res_scale, font_scale = 25, 10, 1500, 0.8
        max_dim = max(self.video_meta_data["width"], self.video_meta_data["height"])
        self.draw_scale, self.font_size = int(
            radius_scale / (res_scale / max_dim)
        ), float(font_scale / (res_scale / max_dim))
        self.spacing_scaler = int(self.space_scale / (res_scale / max_dim))
        self.font = cv2.FONT_HERSHEY_TRIPLEX
        self.__read_roi_dfs()

    def __update_video_meta_data(self):
        new_cap = cv2.VideoCapture(self.video_path)
        new_cap.set(1, 1)
        _, img = self.cap.read()
        bordered_img = cv2.copyMakeBorder(
            img,
            0,
            0,
            0,
            int(self.video_meta_data["width"]),
            borderType=cv2.BORDER_CONSTANT,
            value=[0, 0, 0],
        )
        self.border_img_h, self.border_img_w = (
            bordered_img.shape[0],
            bordered_img.shape[1],
        )
        new_cap.release()

    def __read_roi_dfs(self):
        if not os.path.isfile(
            os.path.join(self.logs_path, "measures", "ROI_definitions.h5")
        ):
            raise NoROIDataError(
                msg="No ROI definitions were found in your SimBA project. Please draw some ROIs before analyzing your ROI data"
            )
        else:
            self.roi_h5_path = os.path.join(
                self.logs_path, "measures", "ROI_definitions.h5"
            )
            self.rectangles_df = pd.read_hdf(self.roi_h5_path, key="rectangles")
            self.circles_df = pd.read_hdf(self.roi_h5_path, key="circleDf")
            self.polygon_df = pd.read_hdf(self.roi_h5_path, key="polygons")
            self.shape_names = list(
                itertools.chain(
                    self.rectangles_df["Name"].unique(),
                    self.circles_df["Name"].unique(),
                    self.polygon_df["Name"].unique(),
                )
            )
            self.video_recs = self.rectangles_df.loc[
                (self.rectangles_df["Video"] == self.video_name)
                & (self.rectangles_df["Name"].isin(self.cue_light_names))
            ]
            self.video_circs = self.circles_df.loc[
                (self.circles_df["Video"] == self.video_name)
                & (self.circles_df["Name"].isin(self.cue_light_names))
            ]
            self.video_polys = self.polygon_df.loc[
                (self.polygon_df["Video"] == self.video_name)
                & (self.polygon_df["Name"].isin(self.cue_light_names))
            ]
            self.shape_names = list(
                itertools.chain(
                    self.rectangles_df["Name"].unique(),
                    self.circles_df["Name"].unique(),
                    self.polygon_df["Name"].unique(),
                )
            )

    def __calc_text_locs(self):
        add_spacer = 2
        self.loc_dict = {}
        for light_cnt, light_name in enumerate(self.cue_light_names):
            self.loc_dict[light_name] = {}
            self.loc_dict[light_name]["status_text"] = "{} {}".format(
                light_name, "status:"
            )
            self.loc_dict[light_name]["onset_cnt_text"] = "{} {}".format(
                light_name, "onset counts:"
            )
            self.loc_dict[light_name]["seconds_on_text"] = "{} {}".format(
                light_name, "time ON (s):"
            )
            self.loc_dict[light_name]["seconds_off_text"] = "{} {}".format(
                light_name, "time OFF (s):"
            )
            self.loc_dict[light_name]["status_text_loc"] = (
                (self.video_meta_data["width"] + 5),
                (
                    self.video_meta_data["height"]
                    - (self.video_meta_data["height"] + 10)
                    + self.spacing_scaler * add_spacer
                ),
            )
            self.loc_dict[light_name]["status_data_loc"] = (
                int(self.border_img_w - (self.border_img_w / 8)),
                (
                    self.video_meta_data["height"]
                    - (self.video_meta_data["height"] + 10)
                    + self.spacing_scaler * add_spacer
                ),
            )
            add_spacer += 1
            self.loc_dict[light_name]["onset_cnt_text_loc"] = (
                (self.video_meta_data["width"] + 5),
                (
                    self.video_meta_data["height"]
                    - (self.video_meta_data["height"] + 10)
                    + self.spacing_scaler * add_spacer
                ),
            )
            self.loc_dict[light_name]["onset_cnt_data_loc"] = (
                int(self.border_img_w - (self.border_img_w / 8)),
                (
                    self.video_meta_data["height"]
                    - (self.video_meta_data["height"] + 10)
                    + self.spacing_scaler * add_spacer
                ),
            )
            add_spacer += 1
            self.loc_dict[light_name]["seconds_on_text_loc"] = (
                (self.video_meta_data["width"] + 5),
                (
                    self.video_meta_data["height"]
                    - (self.video_meta_data["height"] + 10)
                    + self.spacing_scaler * add_spacer
                ),
            )
            self.loc_dict[light_name]["seconds_on_data_loc"] = (
                int(self.border_img_w - (self.border_img_w / 8)),
                (
                    self.video_meta_data["height"]
                    - (self.video_meta_data["height"] + 10)
                    + self.spacing_scaler * add_spacer
                ),
            )
            add_spacer += 1
            self.loc_dict[light_name]["seconds_off_text_loc"] = (
                (self.video_meta_data["width"] + 5),
                (
                    self.video_meta_data["height"]
                    - (self.video_meta_data["height"] + 10)
                    + self.spacing_scaler * add_spacer
                ),
            )
            self.loc_dict[light_name]["seconds_off_data_loc"] = (
                int(self.border_img_w - (self.border_img_w / 8)),
                (
                    self.video_meta_data["height"]
                    - (self.video_meta_data["height"] + 10)
                    + self.spacing_scaler * add_spacer
                ),
            )
            add_spacer += 1

    def __create_text_dict(self):
        self.light_dict = {}
        for light_cnt, light_name in enumerate(self.cue_light_names):
            self.light_dict[light_name] = {}
            self.light_dict[light_name]["status"] = False
            self.light_dict[light_name]["onsets"] = 0
            self.light_dict[light_name]["time_on"] = 0
            self.light_dict[light_name]["time_off"] = 0
            self.light_dict[light_name]["prior_frame_status"] = 0
            self.light_dict[light_name]["color"] = (0, 0, 0)

    def __draw_shapes_and_text(self, shape_data):
        shape_name = shape_data["Name"]
        cv2.putText(
            self.border_img,
            self.loc_dict[shape_name]["status_text"],
            self.loc_dict[shape_name]["status_text_loc"],
            self.font,
            self.font_size,
            shape_data["Color BGR"],
            1,
        )
        cv2.putText(
            self.border_img,
            self.loc_dict[shape_name]["onset_cnt_text"],
            self.loc_dict[shape_name]["onset_cnt_text_loc"],
            self.font,
            self.font_size,
            shape_data["Color BGR"],
            1,
        )
        cv2.putText(
            self.border_img,
            self.loc_dict[shape_name]["seconds_on_text"],
            self.loc_dict[shape_name]["seconds_on_text_loc"],
            self.font,
            self.font_size,
            shape_data["Color BGR"],
            1,
        )
        cv2.putText(
            self.border_img,
            self.loc_dict[shape_name]["seconds_off_text"],
            self.loc_dict[shape_name]["seconds_off_text_loc"],
            self.font,
            self.font_size,
            shape_data["Color BGR"],
            1,
        )
        if shape_data["Shape_type"] == "Rectangle":
            cv2.rectangle(
                self.border_img,
                (shape_data["topLeftX"], shape_data["topLeftY"]),
                (shape_data["Bottom_right_X"], shape_data["Bottom_right_Y"]),
                shape_data["Color BGR"],
                shape_data["Thickness"],
            )
        if shape_data["Shape_type"] == "Circle":
            cv2.circle(
                self.border_img,
                (shape_data["centerX"], shape_data["centerY"]),
                shape_data["radius"],
                shape_data["Color BGR"],
                shape_data["Thickness"],
            )
        if shape_data["Shape_type"] == "Polygon":
            cv2.polylines(
                self.border_img,
                shape_data["vertices"],
                True,
                shape_data["Color BGR"],
                thickness=shape_data["Thickness"],
            )

    def __insert_texts_and_shapes(self):
        for light_cnt, light_name in enumerate(self.cue_light_names):
            for i, r in self.video_recs.iterrows():
                if light_name == r["Name"]:
                    self.__draw_shapes_and_text(shape_data=r)
            for i, r in self.video_circs.iterrows():
                if light_name == r["Name"]:
                    self.__draw_shapes_and_text(shape_data=r)
            for i, r in self.video_polys.iterrows():
                if light_name == r["Name"]:
                    self.__draw_shapes_and_text(shape_data=r)

    def __insert_body_parts(self):
        for animal_name, animal_data in self.animal_bp_dict.items():
            for cnt, (x_bp, y_bp) in enumerate(
                zip(animal_data["X_bps"], animal_data["Y_bps"])
            ):
                cord = tuple(
                    self.data_df.loc[self.frame_cnt, [x_bp, y_bp]].astype(int).values
                )
                cv2.circle(
                    self.border_img,
                    cord,
                    0,
                    animal_data["colors"][cnt],
                    self.draw_scale,
                )

    def run(self):
        """
        Method to create cue light visualizations. Results are stored in the ``project_folder/frames/output/cue_lights``
        directory of the SimBA project.
        """
        self.cap = cv2.VideoCapture(self.video_path)
        self.frame_cnt = 0
        self.__update_video_meta_data()
        if self.video_setting:
            self.fourcc = cv2.VideoWriter_fourcc(*"mp4v")
            self.save_video_path = os.path.join(
                self.output_folder, self.video_name + ".mp4"
            )
            self.writer = cv2.VideoWriter(
                self.save_video_path,
                self.fourcc,
                self.fps,
                (self.border_img_w, self.border_img_h),
            )
        if self.frame_setting:
            self.save_frame_folder_dir = os.path.join(
                self.output_folder, self.video_name
            )
            if not os.path.exists(self.save_frame_folder_dir):
                os.makedirs(self.save_frame_folder_dir)
        self.__calc_text_locs()
        self.__create_text_dict()
        while self.cap.isOpened():
            try:
                _, img = self.cap.read()
                self.border_img = cv2.copyMakeBorder(
                    img,
                    0,
                    0,
                    0,
                    int(self.video_meta_data["width"]),
                    borderType=cv2.BORDER_CONSTANT,
                    value=[0, 0, 0],
                )
                self.border_img_h, self.border_img_w = (
                    self.border_img.shape[0],
                    self.border_img.shape[1],
                )
                self.__insert_texts_and_shapes()
                self.__insert_body_parts()
                for light_name in self.cue_light_names:
                    if (self.light_dict[light_name]["prior_frame_status"] == 0) & (
                        self.data_df.loc[self.frame_cnt, light_name] == 1
                    ):
                        self.light_dict[light_name]["onsets"] += 1
                    if self.data_df.loc[self.frame_cnt, light_name] == 1:
                        self.light_dict[light_name]["color"] = (0, 255, 255)
                        self.light_dict[light_name]["status"] = "ON"
                        self.light_dict[light_name]["time_on"] += 1 / self.fps
                    else:
                        self.light_dict[light_name]["color"] = (90, 10, 10)
                        self.light_dict[light_name]["status"] = "OFF"
                        self.light_dict[light_name]["time_off"] += 1 / self.fps
                    self.light_dict[light_name]["prior_frame_status"] = (
                        self.data_df.loc[self.frame_cnt, light_name]
                    )
                    cv2.putText(
                        self.border_img,
                        str(self.light_dict[light_name]["status"]),
                        self.loc_dict[light_name]["status_data_loc"],
                        self.font,
                        self.font_size,
                        self.light_dict[light_name]["color"],
                        1,
                    )
                    cv2.putText(
                        self.border_img,
                        str(self.light_dict[light_name]["onsets"]),
                        self.loc_dict[light_name]["onset_cnt_data_loc"],
                        self.font,
                        self.font_size,
                        (255, 255, 255),
                        1,
                    )
                    cv2.putText(
                        self.border_img,
                        str(round(self.light_dict[light_name]["time_on"], 2)),
                        self.loc_dict[light_name]["seconds_on_data_loc"],
                        self.font,
                        self.font_size,
                        (255, 255, 255),
                        1,
                    )
                    cv2.putText(
                        self.border_img,
                        str(round(self.light_dict[light_name]["time_off"], 2)),
                        self.loc_dict[light_name]["seconds_off_data_loc"],
                        self.font,
                        self.font_size,
                        (255, 255, 255),
                        1,
                    )
                if self.video_setting:
                    self.writer.write(self.border_img)
                if self.frame_setting:
                    frame_save_path = os.path.join(
                        self.save_frame_folder_dir, str(self.frame_cnt) + ".png"
                    )
                    cv2.imwrite(frame_save_path, self.border_img)
                print(
                    "Cue light frame: {} / {}. Video: {} ".format(
                        str(self.frame_cnt + 1), str(len(self.data_df)), self.video_name
                    )
                )
                self.frame_cnt += 1

            except Exception as e:
                if self.video_setting:
                    self.writer.release()
                print(e.args)
                print(
                    "NOTE: index error / keyerror. Some frames of the video may be missing. Make sure you are running latest version of SimBA with pip install simba-uw-tf-dev"
                )
                break

        if self.video_setting:
            self.writer.release()
        stdout_success(
            msg=f"Cue light visualization for video {self.video_name} saved..."
        )


# test = CueLightVisualizer(config_path='/Users/simon/Desktop/troubleshooting/light_analyzer/project_folder/project_config.ini',
#                           cue_light_names=['Cue_light'],
#                           video_path='/Users/simon/Desktop/troubleshooting/light_analyzer/project_folder/videos/20220422_ALMEAG02_B0.avi',
#                           video_setting=True,
#                           frame_setting=False)
# test.visualize_cue_light_data()



#### MODIFIED FROM @Toshea111 - https://github.com/Toshea111/sleap/blob/develop/docs/notebooks/Convert_HDF5_to_CSV_updated.ipynb
import io
import os

import h5py
import numpy as np
import pandas as pd

from simba.data_processors.interpolation_smoothing import Interpolate, Smooth
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pose_importer_mixin import PoseImporterMixin
from simba.utils.enums import Methods, TagNames
from simba.utils.errors import BodypartColumnNotFoundError
from simba.utils.printing import (SimbaTimer, log_event, stdout_success,
                                  stdout_warning)
from simba.utils.read_write import (clean_sleap_file_name,
                                    find_all_videos_in_project, get_fn_ext,
                                    get_video_meta_data, write_df)


class SLEAPImporterH5(ConfigReader, PoseImporterMixin):
    """
    Importing SLEAP pose-estimation data into SimBA project in ``H5`` format

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str data_folder: Path to folder containing SLEAP data in `csv` format.
    :parameter List[str] id_lst: Animal names. This will be ignored in one animal projects and default to ``Animal_1``.
    :parameter str interpolation_settings: String defining the pose-estimation interpolation method. OPTIONS: 'None', 'Animal(s): Nearest',
        'Animal(s): Linear', 'Animal(s): Quadratic','Body-parts: Nearest', 'Body-parts: Linear',
        'Body-parts: Quadratic'.
    :parameter str smoothing_settings: Dictionary defining the pose estimation smoothing method. EXAMPLE: {'Method': 'Savitzky Golay',
        'Parameters': {'Time_window': '200'}}

    .. note::
       `Multi-animal import tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Multi_animal_pose.md>`__.

    References
    ----------
    .. [1] Pereira et al., SLEAP: A deep learning system for multi-animal pose tracking, `Nature Methods`,
           2022.

    >>> sleap_h5_importer = SLEAPImporterH5(config_path=r'project_folder/project_config.ini', data_folder=r'data_folder', actor_IDs=['Termite_1', 'Termite_2', 'Termite_3', 'Termite_4', 'Termite_5'], interpolation_settings="Body-parts: Nearest", smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
    >>> sleap_h5_importer.run()
    """

    def __init__(
        self,
        config_path: str,
        data_folder: str,
        id_lst: list,
        interpolation_settings: str,
        smoothing_settings: dict,
    ):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        PoseImporterMixin.__init__(self)
        log_event(
            logger_name=str(__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        self.interpolation_settings, self.smoothing_settings = (
            interpolation_settings,
            smoothing_settings,
        )
        self.data_folder, self.id_lst = data_folder, id_lst
        self.import_log_path = os.path.join(
            self.logs_path, f"data_import_log_{self.datetime}.csv"
        )
        self.video_paths = find_all_videos_in_project(videos_dir=self.video_dir)
        self.input_data_paths = self.find_data_files(
            dir=self.data_folder, extensions=[".h5"]
        )
        if self.pose_setting is Methods.USER_DEFINED.value:
            self.__update_config_animal_cnt()
        if self.animal_cnt > 1:
            self.data_and_videos_lk = self.link_video_paths_to_data_paths(
                data_paths=self.input_data_paths,
                video_paths=self.video_paths,
                filename_cleaning_func=clean_sleap_file_name,
            )
            self.check_multi_animal_status()
            self.animal_bp_dict = self.create_body_part_dictionary(
                self.multi_animal_status,
                self.id_lst,
                self.animal_cnt,
                self.x_cols,
                self.y_cols,
                self.p_cols,
                self.clr_lst,
            )
            self.update_bp_headers_file()
        else:
            self.data_and_videos_lk = dict(
                [
                    (get_fn_ext(file_path)[1], {"DATA": file_path, "VIDEO": None})
                    for file_path in self.input_data_paths
                ]
            )
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type="info",
            msg=f"Importing data files: {self.input_data_paths}. Interpolation: {interpolation_settings}, Smoothing: {smoothing_settings}",
        )
        print(f"Importing {len(list(self.data_and_videos_lk.keys()))} file(s)...")

    def run(self):
        for file_cnt, (video_name, video_data) in enumerate(
            self.data_and_videos_lk.items()
        ):
            self.output_filename = clean_sleap_file_name(filename=video_name)
            print(f"Importing {self.output_filename}...")
            video_timer = SimbaTimer(start=True)
            self.video_name = self.output_filename
            with h5py.File(video_data["DATA"], "r") as f:
                missing_keys = [
                    x
                    for x in ["tracks", "point_scores", "node_names", "track_names"]
                    if not x in list(f.keys())
                ]
                if missing_keys:
                    stdout_warning(
                        msg=f'{video_data["DATA"]} is not a valid SLEAP H5 file. Missing keys {missing_keys} Skipping {self.output_filename}...'
                    )
                    continue
                tracks = f["tracks"][:].T
                point_scores = f["point_scores"][:].T

            csv_rows = []
            n_frames, n_nodes, _, n_tracks = tracks.shape
            for frame_ind in range(n_frames):
                csv_row = []
                for track_ind in range(n_tracks):
                    for node_ind in range(n_nodes):
                        for xyp in range(3):
                            if xyp == 0 or xyp == 1:
                                data = tracks[frame_ind, node_ind, xyp, track_ind]
                            else:
                                data = point_scores[frame_ind, node_ind, track_ind]

                            csv_row.append(f"{data:.3f}")
                csv_rows.append(" ".join(csv_row))
            csv_rows = "\n".join(csv_rows)
            self.data_df = pd.read_csv(
                io.StringIO(csv_rows), delim_whitespace=True, header=None
            ).fillna(0)
            if len(self.data_df.columns) != len(self.bp_headers):
                raise BodypartColumnNotFoundError(
                    msg=f'The number of body-parts in data file {video_data["DATA"]} do not match the number of body-parts in your SimBA project. '
                    f"The number of of body-parts expected by your SimBA project is {int(len(self.bp_headers) / 3)}. "
                    f'The number of of body-parts contained in data file {video_data["DATA"]} is {int(len(self.data_df.columns) / 3)}. '
                    f"Make sure you have specified the correct number of animals and body-parts in your project.",
                    source=self.__class__.__name__,
                )
            self.data_df.columns = self.bp_headers
            if self.animal_cnt > 1:
                self.initialize_multi_animal_ui(
                    animal_bp_dict=self.animal_bp_dict,
                    video_info=get_video_meta_data(video_data["VIDEO"]),
                    data_df=self.data_df,
                    video_path=video_data["VIDEO"],
                )
                self.multianimal_identification()

            else:
                self.out_df = self.insert_multi_idx_columns(df=self.data_df.fillna(0))

            self.save_path = os.path.join(
                os.path.join(
                    self.input_csv_dir, f"{self.output_filename}.{self.file_type}"
                )
            )
            write_df(
                df=self.out_df,
                file_type=self.file_type,
                save_path=self.save_path,
                multi_idx_header=True,
            )
            if self.interpolation_settings != "None":
                self.__run_interpolation()
            if self.smoothing_settings["Method"] != "None":
                self.__run_smoothing()
            video_timer.stop_timer()
            stdout_success(
                msg=f"Video {self.output_filename} data imported...",
                elapsed_time=video_timer.elapsed_time_str,
                source=self.__class__.__name__,
            )
        self.timer.stop_timer()
        stdout_success(
            msg="All SLEAP H5 data files imported",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )

    def __run_interpolation(self):
        print(
            f"Interpolating missing values in video {self.output_filename} (Method: {self.interpolation_settings})..."
        )
        _ = Interpolate(
            input_path=self.save_path,
            config_path=self.config_path,
            method=self.interpolation_settings,
            initial_import_multi_index=True,
        )

    def __run_smoothing(self):
        print(
            f'Performing {self.smoothing_settings["Method"]} smoothing on video {self.output_filename}...'
        )
        Smooth(
            config_path=self.config_path,
            input_path=self.save_path,
            time_window=int(self.smoothing_settings["Parameters"]["Time_window"]),
            smoothing_method=self.smoothing_settings["Method"],
            initial_import_multi_index=True,
        )


# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/sleap_import/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_import/data_h5',
#                    id_lst=['White', 'Black'],
#                    interpolation_settings= "Body-parts: Nearest", #'"Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'None', 'Parameters': {'Time_window': '200'}})
# test.run()


# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/sleap_dropped_frms_2/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_dropped_frms_2/data',
#                    id_lst=['Simon'],
#                    interpolation_settings= 'None', #'"Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'None', 'Parameters': {'Time_window': '200'}})
# test.run()

# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/import_h5',
#                    id_lst=['Simon', 'Nastacia', 'Ben', 'John', 'JJ'],
#                    interpolation_settings="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
# test.run()
# print('All SLEAP imports complete.')


#
# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/sleap_dropped_frms/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_dropped_frms/data',
#                    id_lst=['Simon'],
#                    interpolation_settings= 'None', #'"Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'None', 'Parameters': {'Time_window': '200'}})
# test.run()


# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/SLEAP_2_Animals_16_body_parts/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/SLEAP_2_Animals_16_body_parts/data/h5',
#                    id_lst=['Simon', 'Nastacia'],
#                    interpolation_settings="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
# test.run()
# print('All SLEAP imports complete.')


# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/import_h5',
#                    id_lst=['Simon', 'Nastacia', 'Ben', 'John', 'JJ'],
#                    interpolation_settings='None',
#                    smoothing_settings = {'Method': 'None'})
# test.run()
# print('All SLEAP imports complete.')
#
# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/sleap_cody/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_cody/import_h5',
#                    actor_IDs=['Nastacia', 'Sam'],
#                    interpolation_settings="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
# test.import_sleap()
# print('All SLEAP imports complete.')


# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/sleap_cody/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_cody/import_h5',
#                    actor_IDs=['Nastacia'],
#                    interpolation_settings="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
# test.import_sleap()
# # print('All SLEAP imports complete.')



import multiprocessing
import os
import shutil
import subprocess
from typing import Optional, Union

from simba.mixins.video_processing_mixin import VideoProcessingMixin
from simba.utils.checks import check_file_exist_and_readable, check_int
from simba.utils.data import re
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder, get_fn_ext,
                                    get_video_meta_data)
from simba.utils.warnings import FileExistWarning, SameInputAndOutputWarning


def change_single_video_fps(
    file_path: Union[str, os.PathLike],
    fps: int,
    gpu: Optional[bool] = False,
    core_cnt: Optional[bool] = 1,
) -> None:
    """
    Change the fps of a single video file. Results are stored in the same directory as in the input file with
    the suffix ``_fps_new_fps``.

    :parameter Union[str, os.PathLike] file_path: Path to video file
    :parameter int fps: Fps of the new video file.

    :example:
    >>> _ = change_single_video_fps(file_path='project_folder/videos/Video_1.mp4', fps=15)
    """

    check_file_exist_and_readable(file_path=file_path)
    check_int(name="New fps", value=fps)
    check_int(name="CORE COUNT", value=core_cnt)
    video_processor = VideoProcessingMixin()
    video_meta_data = get_video_meta_data(video_path=file_path)
    dir_name, file_name, ext = get_fn_ext(filepath=file_path)
    if int(fps) == int(video_meta_data["fps"]):
        SameInputAndOutputWarning(
            msg=f"The new fps is the same as the input fps for video {file_name} ({str(fps)})"
        )
    save_path = os.path.join(
        dir_name, file_name + "_fps_{}{}".format(str(fps), str(ext))
    )
    if os.path.isfile(save_path):
        FileExistWarning(msg=f"Overwriting existing file at {save_path}")

    timer = SimbaTimer(start=True)
    if core_cnt > 1:
        video_input_paths = video_processor.split_video_into_n_cunks(
            video_path=file_path, n=core_cnt
        )
        video_output_paths = video_processor.create_ouput_paths(
            video_input_paths=video_input_paths
        )
        cmds = []
        for input_path, output_path in zip(video_input_paths, video_output_paths):
            if gpu:
                cmds.append(
                    'ffmpeg -hwaccel auto -i "{}" -filter:v fps=fps={} "{}"'.format(
                        input_path, fps, output_path
                    )
                )
            else:
                cmds.append(
                    'ffmpeg -i "{}" -filter:v fps=fps={} "{}" -threads 8'.format(
                        input_path, fps, output_path
                    )
                )
        cmds = [x for x in cmds]
        with multiprocessing.Pool(core_cnt) as pool:
            for cnt, result in enumerate(
                pool.map(video_processor._ffmpeg_cmd_multiprocessor, cmds, chunksize=1)
            ):
                print(
                    f"Video {file_name} {round((((cnt+1)/len(cmds)) * 100), 2)}% complete..."
                )
            pool.terminate()
            pool.join()
    #     concatenate_videos_in_folder(in_folder=os.path.dirname(video_output_paths[0]), save_path=save_path, remove_splits=True)
    #     shutil.rmtree(os.path.dirname(video_input_paths[0]))
    else:
        command = (
            str("ffmpeg -i ")
            + '"'
            + str(file_path)
            + '"'
            + " -filter:v fps=fps="
            + str(fps)
            + " "
            + '"'
            + save_path
            + '" -y'
        )
        subprocess.call(command, shell=True)

    timer.stop_timer()
    stdout_success(
        msg=f'SIMBA COMPLETE: FPS of video {file_name} changed from {str(video_meta_data["fps"])} to {str(fps)} and saved in directory {save_path}',
        elapsed_time=timer.elapsed_time_str,
    )

    #
    #
    # else:
    #     cmd =
    #
    # with multiprocessing.Pool(core_cnt, maxtasksperchild=10) as pool:
    #     video_paths = pool.map(video_processor._ffmpeg_cmd_multiprocessor, video_path)


change_single_video_fps(
    file_path="/Users/simon/Downloads/A_Ko15 8 PushOff.avi",
    fps=2,
    gpu=False,
    core_cnt=1,
)



__author__ = "Simon Nilsson"

import glob
import os
import shutil
from datetime import datetime

import pandas as pd

from simba.feature_extractors.feature_extractor_4bp import \
    ExtractFeaturesFrom4bps
from simba.feature_extractors.feature_extractor_7bp import \
    ExtractFeaturesFrom7bps
from simba.feature_extractors.feature_extractor_8bp import \
    ExtractFeaturesFrom8bps
from simba.feature_extractors.feature_extractor_9bp import \
    ExtractFeaturesFrom9bps
from simba.feature_extractors.feature_extractor_14bp import \
    ExtractFeaturesFrom14bps
from simba.feature_extractors.feature_extractor_16bp import \
    ExtractFeaturesFrom16bps
from simba.feature_extractors.feature_extractor_user_defined import \
    UserDefinedFeatureExtractor
from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_file_exist_and_readable,
                                check_that_column_exist)
from simba.utils.read_write import (get_all_clf_names, get_fn_ext,
                                    read_config_entry, read_df, write_df)


class Reverse2AnimalTracking(ConfigReader):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        self.in_dir = os.path.join(
            self.project_path, "csv", "outlier_corrected_movement_location"
        )
        self.features_path = os.path.join(
            self.project_path, "csv", "features_extracted"
        )
        self.targets_path = os.path.join(self.project_path, "csv", "targets_inserted")
        self.model_cnt = read_config_entry(
            self.config, "SML settings", "No_targets", data_type="int"
        )
        self.store_path_features = os.path.join(
            self.features_path, "Non_reversed_files_at_" + str(self.datetime)
        )
        self.store_path_targets = os.path.join(
            self.targets_path, "Non_reversed_files_at_" + str(datetime)
        )
        self.store_path_outliers = os.path.join(
            self.in_dir, "Non_reversed_files_at_" + str(datetime)
        )
        if not os.path.exists(self.store_path_features):
            os.makedirs(self.store_path_features)
        if not os.path.exists(self.store_path_targets):
            os.makedirs(self.store_path_targets)
        if not os.path.exists(self.store_path_outliers):
            os.makedirs(self.store_path_outliers)
        self.pose_estimation_setting = read_config_entry(
            self.config, "create ensemble settings", "pose_estimation_body_parts", "str"
        )

        self.files_found = glob.glob(self.in_dir + "/*." + self.file_type)
        print("Processing {} video(s)...".format(str(len(self.files_found))))

    def reverse_tracking(self):
        for file_cnt, file_path in enumerate(self.files_found):
            _, self.video_name, _ = get_fn_ext(file_path)
            self.data_df = read_df(file_path, self.file_type)
            self.reversed_df = pd.DataFrame()
            animal_dict, col_counter = {}, 0
            for animal_cnt, (animal_name, animal_bps) in enumerate(
                self.animal_bp_dict.items()
            ):
                animal_dict[animal_cnt] = self.data_df.iloc[
                    :, col_counter : col_counter + len(animal_bps["X_bps"] * 3)
                ]
                col_counter += len(animal_bps["X_bps"] * 3)
            for k in reversed(list(animal_dict.keys())):
                self.reversed_df = pd.concat([self.reversed_df, animal_dict[k]], axis=1)
            shutil.move(
                file_path,
                os.path.join(self.store_path_outliers, os.path.basename(file_path)),
            )
            write_df(self.reversed_df, self.file_type, file_path)

    def create_features(self):
        old_feature_files = glob.glob(self.features_path + "/*." + self.file_type)
        for file_path in old_feature_files:
            shutil.move(
                file_path,
                os.path.join(self.store_path_features, os.path.basename(file_path)),
            )
        if self.pose_estimation_setting == "16":
            ExtractFeaturesFrom16bps(self.config_path)
        elif self.pose_estimation_setting == "14":
            ExtractFeaturesFrom14bps(self.config_path)
        elif self.pose_estimation_setting == "987":
            extract_features_wotarget_14_from_16(self.config_path)
        elif self.pose_estimation_setting == "9":
            extract_features_wotarget_9(self.config_path)
        elif self.pose_estimation_setting == "8":
            ExtractFeaturesFrom8bps(self.config_path)
        elif self.pose_estimation_setting == "7":
            ExtractFeaturesFrom7bps(self.config_path)
        elif self.pose_estimation_setting == "4":
            ExtractFeaturesFrom4bps(self.config_path)
        elif self.pose_estimation_setting == "user_defined":
            UserDefinedFeatureExtractor(self.config_path)

    def reappend_targets(self):
        clf_names = get_all_clf_names(config=self.config, target_cnt=self.model_cnt)
        feature_files = glob.glob(self.features_path + "/*." + self.file_type)
        for file_path in feature_files:
            _, video_name, _ = get_fn_ext(file_path)
            data_df = read_df(file_path, self.file_type)
            target_df_path = (
                os.path.join(self.targets_path, video_name) + "." + self.file_type
            )
            check_file_exist_and_readable(target_df_path)
            target_df = read_df(target_df_path, self.file_type)
            for clf_name in clf_names:
                check_that_column_exist(
                    target_df, column_name=clf_name, file_name=file_path
                )
                check_that_two_dfs_are_equal_len(
                    df_1=data_df[clf_name],
                    df_2=target_df[clf_name],
                    file_path_1=file_path,
                    file_path_2=target_df_path,
                    col_name=clf_name,
                )
                data_df[clf_name] = target_df[clf_name]
            shutil.move(
                file_path,
                os.path.join(self.store_path_targets, os.path.basename(file_path)),
            )


test = Reverse2AnimalTracking(
    config_path="/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini"
)
test.reverse_tracking()



### CODE MODIFID FROM PYBURST PACKAGE  https://pypi.org/project/pybursts/
import math

import numpy as np


def kleinberg_burst_detection(offsets: np.ndarray, s: float, gamma: float):
    """
    Burst detection using `pyburst <https://pypi.org/project/pybursts/>`_.
    Private method called by ``simba.data_processors.kleinberg_calculator.KleinbergCalculator``.

    """

    offsets = np.array(offsets, dtype=object)

    if offsets.size == 1:
        bursts = np.array([0, offsets[0], offsets[0]], ndmin=2, dtype=object)
        return bursts

    offsets = np.sort(offsets)
    gaps = np.diff(offsets)

    if not np.all(gaps):
        raise ValueError("Input cannot contain events with zero time between!")

    T = np.sum(gaps)
    n = np.size(gaps)

    g_hat = T / n

    k = int(math.ceil(float(1 + math.log(T, s) + math.log(1 / np.amin(gaps), s))))

    gamma_log_n = gamma * math.log(n)

    def tau(i, j):
        if i >= j:
            return 0
        else:
            return (j - i) * gamma_log_n

    alpha_function = np.vectorize(lambda x: s**x / g_hat)
    alpha = alpha_function(np.arange(k))

    def f(j, x):
        return alpha[j] * math.exp(-alpha[j] * x)

    C = np.repeat(float("inf"), k)
    C[0] = 0

    q = np.empty((k, 0))
    for t in range(n):
        C_prime = np.repeat(float("inf"), k)
        q_prime = np.empty((k, t + 1))
        q_prime.fill(np.nan)

        for j in range(k):
            cost_function = np.vectorize(lambda x: C[x] + tau(x, j))
            cost = cost_function(np.arange(0, k))

            el = np.argmin(cost)

            if f(j, gaps[t]) > 0:
                C_prime[j] = cost[el] - math.log(f(j, gaps[t]))

            if t > 0:
                q_prime[j, :t] = q[el, :]

            q_prime[j, t] = j + 1

        C = C_prime
        q = q_prime

    j = np.argmin(C)
    q = q[j, :]

    prev_q = 0

    N = 0
    for t in range(n):
        if q[t] > prev_q:
            N = N + q[t] - prev_q
        prev_q = q[t]

    bursts = np.array(
        [np.repeat(np.nan, N), np.repeat(offsets[0], N), np.repeat(offsets[0], N)],
        ndmin=2,
        dtype=object,
    ).transpose()
    burst_counter = -1
    prev_q = 0
    stack = np.repeat(np.nan, N)
    stack_counter = -1
    for t in range(n):
        if q[t] > prev_q:
            num_levels_opened = q[t] - prev_q
            for i in range(int(num_levels_opened)):
                burst_counter += 1
                bursts[burst_counter, 0] = prev_q + i
                bursts[burst_counter, 1] = offsets[t]
                stack_counter += 1
                stack[stack_counter] = burst_counter
        elif q[t] < prev_q:
            num_levels_closed = prev_q - q[t]
            for i in range(int(num_levels_closed)):
                bursts[int(stack[stack_counter]), 2] = offsets[t]
                stack_counter -= 1
        prev_q = q[t]

    while stack_counter >= 0:
        bursts[int(stack[stack_counter]), 2] = offsets[n]
        stack_counter -= 1

    return bursts
